from .depthcharge.encoders import PeakEncoder, FloatEncoder
from .depthcharge.transformers import SpectrumTransformerEncoder, PeptideTransformerDecoder
from typing import Any, Dict, Iterable, List, Tuple
import torch
import numpy as np
import pytorch_lightning as pl

class AugmentedPeakEncoder(torch.nn.Module):
    """Encode an augmented m/z, intensity,

    Parameters
    ----------
    d_model : int
        The number of features to output.
    min_rt_wavelength : float, optional
        The minimum wavelength to use for m/z.
    max_rt_wavelength : float, optional
        The maximum wavelength to use for m/z.
    min_intensity_wavelength : float, optional
        The minimum wavelength to use for intensity. The default assumes
        intensities between [0, 1].
    max_intensity_wavelength : float, optional
        The maximum wavelength to use for intensity. The default assumes
        intensities between [0, 1].
    """

    def __init__(
        self,
        d_model: int,
        min_intensity_wavelength: float = 1e-6,
        max_intensity_wavelength: float = 1,
        min_rt_wavelength: float = 1e-6,
        max_rt_wavelength: float = 10
    ) -> None:
        """Initialize the MzEncoder."""
        super().__init__()
        self.d_model = d_model

        self.peak_encoder = PeakEncoder(
            d_model,
            min_intensity_wavelength,
            max_intensity_wavelength,
            learnable_wavelengths = False
        )

        self.rt_encoder = FloatEncoder(
            d_model,
            min_wavelength=min_rt_wavelength,
            max_wavelength=max_rt_wavelength
        )

        self.level_encoder = torch.nn.Embedding(3, d_model)

        self.combiner = torch.nn.Linear(3 * d_model, d_model, bias=False)

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """Encode m/z values, intensities, rts, and mslevel

        Note that we expect intensities to fall within the interval [0, 1].

        Parameters
        ----------
        X : torch.Tensor of shape (n_spectra, n_peaks, 4)
            The spectra to embed. Axis 0 represents a mass spectrum, axis 1
            contains the peaks in the mass spectrum, and axis 2 is a 4-tuple
            specifying the (m/z, intensity, retention time, ms level) for each peak.
            These are zero-padded, such that all of the spectra in the batch
            are the same length.

        Returns
        -------
        torch.Tensor of shape (n_spectra, n_peaks, d_model)
            The encoded features for the augmented mass spectra.
        """

        encoded = torch.cat(
            [
                self.peak_encoder(X),
                self.rt_encoder(X[:, :, 2]),
                self.level_encoder(X[:, :, 3].int())
            ],
            dim=2,
        )

        return self.combiner(encoded)

class AugmentedSpec2Pep(pl.LightningModule):
    def __init__(
            self,
            d_model,
            n_layers,
            rt_width,
            n_head,
            dropout,
            dim_feedforward,
            tokenizer,
            max_charge,
            lr = 1e-4,
            frag_weight = 20,
            lr_decay=1e-9, 
            alpha = 1e-3
        ):

        super().__init__()

        self.peak_encoder = AugmentedPeakEncoder(
            d_model=d_model,
            max_rt_wavelength=2*rt_width
        )

        self.spectrum_encoder = SpectrumTransformerEncoder(
            d_model=d_model,
            n_head=n_head,
            n_layers=n_layers,
            dropout=dropout,
            dim_feedforward=dim_feedforward,
            peak_encoder=self.peak_encoder
        )

        self.decoder = PeptideTransformerDecoder(
            d_model=d_model,
            nhead=n_head,
            dim_feedforward=dim_feedforward,
            n_layers=n_layers,
            dropout=dropout,
            n_tokens=tokenizer,
            max_charge=max_charge
        )

        self.prec_layer = torch.nn.Linear(d_model,1)
        self.frag_layer = torch.nn.Linear(d_model,2)

        self.CELoss = torch.nn.CrossEntropyLoss(ignore_index=0)
        self.fragCELoss = torch.nn.CrossEntropyLoss(weight=torch.tensor([1.,frag_weight]))
        self.MSELoss = torch.nn.MSELoss()
        self.lr = lr
        self.lr_decay=lr_decay
        self.alpha = alpha
        self.tokenizer = tokenizer

    def _forward_step(
        self,
        spectra: torch.Tensor,
        precursors: torch.Tensor,
        sequences: List[str],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        emb = self.spectrum_encoder(spectra)
        result =  self.decoder(sequences, precursors, *emb)
        spec_rep = emb[0][:,0]
        pred_prec = self.prec_layer(spec_rep)[:,0]
        pred_frag = self.frag_layer(emb[0])

        return result, pred_prec, pred_frag

    def training_step(self, batch):

        spectra, precursors, sequences, frag_labels, _ = batch
        preds, pred_prec, pred_frags= self._forward_step(spectra, precursors, sequences)
        preds_seqs = torch.argmax(preds, dim=2)[:,:-1]
        preds = preds[:,:-1,:].reshape(-1, len(self.tokenizer) + 1)
        true_seqs = batch[2][:,:]

        loss = self.CELoss(preds,true_seqs.flatten())

        true_prec = batch[1][:,0]/1000
        mse_loss = self.MSELoss(pred_prec, true_prec)

        pred_frags = pred_frags[:,1:,:].reshape(-1, 2)
        frag_loss = self.alpha * self.fragCELoss(pred_frags, frag_labels)

        correct_aas = torch.logical_or(preds_seqs == true_seqs, true_seqs == 0)
        aa_acc = torch.mean(correct_aas.float())

        correct_peps = torch.mean(correct_aas.float(), dim=1) == 1
        pep_acc = torch.mean(correct_peps.float())

        frag_acc = torch.mean((torch.argmax(pred_frags, dim=1) == frag_labels).float())
        frag_baseline = 1 - torch.sum(frag_labels)/len(frag_labels)

        self.log(
            "Train CELoss",
            loss.detach()
        )
        self.log(
            "Train Frag Loss",
            frag_loss.detach()
        )
        self.log(
            "Train AA Acc.",
            aa_acc.detach()
        )
        self.log(
            "Train Pep. Acc.",
            pep_acc.detach()
        )
        self.log(
            "Train Frag Acc.",
            frag_acc.detach()
        )

        return loss + frag_loss

    def validation_step(self, batch, _):
        torch.set_grad_enabled(True)

        spectra, precursors, sequences, frag_labels, _ = batch
        preds, pred_prec, pred_frags = self._forward_step(spectra, precursors, sequences)
        preds_seqs = torch.argmax(preds, dim=2)[:,:-1]
        preds = preds[:,:-1,:].reshape(-1, len(self.tokenizer) + 1)
        true_seqs = batch[2][:,:]
        loss = self.CELoss(preds,true_seqs.flatten())

        pred_frags = pred_frags[:,1:,:].reshape(-1, 2)
        frag_loss = self.alpha*self.fragCELoss(pred_frags, frag_labels)
        frag_acc = torch.mean((torch.argmax(pred_frags, dim=1) == frag_labels).float())
        frag_baseline = 1 - torch.sum(frag_labels)/len(frag_labels)

        correct_aas = torch.logical_or(preds_seqs == true_seqs, true_seqs == 0)
        padding = torch.sum(true_seqs == 0)
        aa_acc = (torch.sum(correct_aas.float()) - padding) / torch.sum(true_seqs != 0)

        correct_peps = torch.mean(correct_aas.float(), dim=1) == 1
        pep_acc = torch.mean(correct_peps.float())

        self.log(
            f"Val CELoss",
            loss.detach(),
            on_epoch=True,
            sync_dist=True
        )
        self.log(
            f"Val Frag Loss",
            frag_loss.detach(),
            on_epoch=True,
            sync_dist=True
        )
        self.log(
            f"Val AA Acc.",
            aa_acc.detach(),
            on_epoch=True,
            sync_dist=True
        )
        self.log(
            f"Val Pep. Acc.",
            pep_acc.detach(),
            on_epoch=True,
            sync_dist=True
        )
        self.log(
            f"Val Frag Acc.",
            frag_acc.detach(),
            on_epoch=True,
            sync_dist=True
        )

        return loss + frag_loss, pep_acc

    def predict_step(self, batch, *args):
        torch.set_grad_enabled(True)

        spectra, precursors, sequences, frag_labels, rts = batch
        device = spectra.get_device()
        if device < 0:
            device = 'cpu'

        cur_sequences = torch.empty((len(sequences),0), dtype=torch.int32, device=device)
        aa_conf = torch.empty((len(sequences),0), dtype=torch.int32, device=device)
        for i in range(len(sequences[0]) + 1):
            preds, _, _ = self._forward_step(spectra, precursors, cur_sequences)
            next_aa_scores = torch.softmax(preds[:,-1,:], 1)
            next_aas = torch.argmax(next_aa_scores, 1)
            aa_conf = torch.cat([aa_conf, torch.reshape(next_aa_scores[range(len(next_aas)),next_aas], (-1, 1))], dim=1)
            cur_sequences = torch.cat([cur_sequences, torch.reshape(next_aas, (-1, 1))], dim=1)

        preds_seqs = cur_sequences
        true_seqs = batch[2]

        pred_seqs = self.tokenizer.detokenize(preds_seqs)
        true_seqs = self.tokenizer.detokenize(true_seqs)

        pep_conf = []
        trimmed_pred_seqs = []
        for pred_pep, pep_aa_conf in zip(pred_seqs, aa_conf):
            trimmed_pep = pred_pep.split('$')[0]
            trimmed_pred_seqs.append(trimmed_pep)
            pep_conf.append(torch.sum(torch.log(pep_aa_conf[:len(trimmed_pep) + 1])))

        return trimmed_pred_seqs, true_seqs, pep_conf, aa_conf, rts, precursors

    def configure_optimizers(self):
        """
        Initialize the optimizer.
        """
        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr, weight_decay = self.lr_decay)
        # Apply learning rate scheduler per step.
        lr_scheduler = CosineWarmupScheduler(
            optimizer, warmup=10000, max_iters=100000
        )
        return [optimizer], {"scheduler": lr_scheduler, "interval": "step"}


class CosineWarmupScheduler(torch.optim.lr_scheduler._LRScheduler):
    """
    Learning rate scheduler with linear warm up followed by cosine shaped decay.

    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        Optimizer object.
    warmup : int
        The number of warm up iterations.
    max_iters : torch.optim
        The total number of iterations.
    """

    def __init__(
        self, optimizer: torch.optim.Optimizer, warmup: int, max_iters: int
    ):
        self.warmup, self.max_iters = warmup, max_iters
        super().__init__(optimizer)

    def get_lr(self):
        lr_factor = self.get_lr_factor(epoch=self.last_epoch)
        return [base_lr * lr_factor for base_lr in self.base_lrs]

    def get_lr_factor(self, epoch):
        lr_factor = 0.5 * (1 + np.cos(np.pi * epoch / self.max_iters))
        if epoch <= self.warmup:
            lr_factor *= epoch / self.warmup
        return lr_factor



"""Datasets for working with peptide sequences."""
from collections.abc import Iterable

import torch
from torch.utils.data import DataLoader, TensorDataset

from ..tokenizers import PeptideTokenizer


class PeptideDataset(TensorDataset):
    """A dataset for peptide sequences.

    Parameters
    ----------
    tokenizer : PeptideTokenizer
        A tokenizer specifying how to transform peptide sequences.
        into tokens.
    sequences : Iterable[str]
        The peptide sequences in a format compatible with
        your tokenizer. ProForma is preferred.
    charges : torch.Tensor,
        The charge state for each peptide.
    *args : torch.Tensor, optional
        Additional values to include during data loading.
    """

    def __init__(
        self,
        tokenizer: PeptideTokenizer,
        sequences: Iterable[str],
        charges: torch.Tensor,
        *args: torch.Tensor,
    ) -> None:
        """Initialize a PeptideDataset."""
        tokens = tokenizer.tokenize(sequences)
        super().__init__(tokens, charges, *args)

    @property
    def tokens(self) -> torch.Tensor:
        """The peptide sequence tokens."""
        return self.tensors[0]

    @property
    def charges(self) -> torch.Tensor:
        """The peptide charges."""
        return self.tensors[1]

    def loader(self, *args: tuple, **kwargs: dict) -> DataLoader:
        """A PyTorch DataLoader for peptides.

        Parameters
        ----------
        *args : tuple
            Arguments passed initialize a torch.utils.data.DataLoader,
            excluding ``dataset``.
        **kwargs : dict
            Keyword arguments passed initialize a torch.utils.data.DataLoader,
            excluding ``dataset``.

        Returns
        -------
        torch.utils.data.DataLoader
            A DataLoader for the peptide.
        """
        return DataLoader(self, *args, **kwargs)



"""A flexible feed-forward neural network."""
from collections.abc import Iterable

import numpy as np
import torch


class FeedForward(torch.nn.Module):
    """Create a feed forward neural network.

    Parameters
    ----------
    in_features : int
        The input dimensionality.
    out_features : int
        The output dimensionality.
    layers : int or tuple of int.
        If an int, layer sizes are linearly interpolated between the input and
        output dimensions using this number of layers. Otherwise, each element
        specifies the size of a layer.
    dropout : float, optionalf
        If greater than zero, add dropout layers with the specified
        probability.
    activation: torch.nn.Module, optional
        The activation function to place between layers.
    append : torch.nn.Module or None, optional
        A final layer to append, such as a sigmoid or tanh.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        layers: int | Iterable[int],
        dropout: float = 0,
        activation: torch.nn.Module = torch.nn.LeakyReLU(),
        append: torch.nn.Module | None = None,
    ) -> None:
        """Initiazlize a FeedForward network."""
        super().__init__()
        try:
            sizes = np.array(
                [in_features] + list(layers) + [out_features],
            )
        except TypeError:
            sizes = np.ceil(
                np.linspace(in_features, out_features, int(layers) + 1),
            )

        sizes = sizes.astype(int)
        stack = []
        for idx in range(len(sizes) - 1):
            stack.append(torch.nn.Linear(sizes[idx], sizes[idx + 1]))
            if idx < len(sizes) - 2:
                stack.append(activation)
                if dropout:
                    stack.append(torch.nn.Dropout(dropout))

        if append is not None:
            stack.append(append)

        self.layers = torch.nn.Sequential(*stack)

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """The forward pass.

        Parameters
        ----------
        X : torch.Tensor of shape (..., in_features)
            The input tensor.

        Returns
        -------
        torch.Tensor of shape (..., out_features)
            The output tensor.
        """
        return self.layers(X)



"""Fundamental dataclasses for depthcharge."""
from __future__ import annotations

import re
from collections.abc import Sequence
from dataclasses import dataclass

import numpy as np
import selfies as sf
import torch
from numpy.typing import ArrayLike
from PIL.PngImagePlugin import PngImageFile
from pyteomics import proforma
from pyteomics.proforma import GenericModification, MassModification
from rdkit import Chem
from rdkit.Chem import Draw
from spectrum_utils.spectrum import MsmsSpectrum
from pyteomics import mass

from .constants import PROTON

MSKB_TO_UNIMOD = {
    "+42.011": "[Acetyl]-",
    "+43.006": "[Carbamyl]-",
    "-17.027": "[Ammonia-loss]-",
    "+43.006-17.027": "[+25.980265]-",  # Not in Unimod
    "M+15.995": "M[Oxidation]",
    "N+0.984": "N[Deamidated]",
    "Q+0.984": "Q[Deamidated]",
    "C+57.021": "C[Carbamidomethyl]",
}


@dataclass
class Peptide:
    """A peptide sequence with or without modification and/or charge.

    Parameters
    ----------
    sequence : str
        The bare amino acid sequence.
    modifications : iterable of str, float, or None, optional
        The modification at each amino acid. This should be the length of
        ``sequence`` + 2, where index 0 and -1 are used for N- and C-terminal
        modification respectively. Use ``None`` in the iterable to specify
        unmodified positions. When ``modifications`` is ``None``, it is assumed
        that no modifications are present.
    charge : int, optional
        The charge of the peptide.
    """

    sequence: str
    modifications: Sequence[str | float | None] | None = None
    charge: int | None = None

    def __post_init__(self) -> None:
        """Validate the provided parameters."""
        if self.modifications is not None:
            if len(self.modifications) != len(self.sequence) + 2:
                raise ValueError(
                    "'modifications' must be two elements longer than "
                    "'sequences' to account for terminal modifications."
                )

        if self.modifications is None:
            self.modifications = [None] * (len(self.sequence) + 2)

        # Parse modifications into Pyteomics' format
        parsed = [None] * (len(self.sequence) + 2)
        for idx, mod in enumerate(self.modifications):
            if mod is None:
                continue

            try:
                mod = [MassModification(mod)]
            except ValueError:
                try:
                    mod = [GenericModification(mod)]
                except (AttributeError, TypeError):
                    pass
            except TypeError:
                pass

            parsed[idx] = mod

        self.modifications = parsed
        n_mod = self.modifications[0]
        c_mod = self.modifications[-1]

        self.proforma = proforma.to_proforma(
            sequence=list(zip(self.sequence, self.modifications[1:-1])),
            n_term=[n_mod] if n_mod is not None else n_mod,
            c_term=[c_mod] if c_mod is not None else c_mod,
            charge_state=self.charge,
        )

    def split(self) -> list[str]:
        """Split the modified peptide for tokenization."""
        if self.modifications is None:
            return list(self.sequence)

        out = []
        for idx, (aa, mods) in enumerate(
            zip(f"-{self.sequence}-", self.modifications)
        ):
            if mods is None:
                if idx and (idx < len(self.modifications) - 1):
                    out.append(aa)

                continue

            if len(mods) == 1:
                try:
                    modstr = f"[{mods[0].name}]"
                except (AttributeError, ValueError):
                    modstr = f"[{mods[0].mass:+0.6f}]"
            else:
                modstr = f"[{sum([m.mass for m in mods]):+0.6f}]"

            if not idx:
                out.append(f"{modstr}-")
            else:
                out.append(f"{aa}{modstr}")

        return out

    @classmethod
    def from_proforma(
        cls,
        sequence: str,
    ) -> Peptide:
        """Create a Peptide from a ProForma 2.0 string.

        Parameters
        ----------
        sequence : str
            A ProForma 2.0-compliant string.

        Returns
        -------
        Peptide
            The parsed ProForma peptide.
        """
        pep, meta = proforma.parse(sequence)
        try:
            charge = meta["charge_state"].charge
        except AttributeError:
            charge = None

        static_mods = {}
        for mod_rule in meta["fixed_modifications"]:
            for res in mod_rule.targets:
                static_mods[res] = mod_rule.modification_tag

        seq = [None] * len(pep)
        mods = [None] * (len(pep) + 2)
        mods[0] = meta["n_term"]
        mods[-1] = meta["c_term"]
        for idx, (res, var_mods) in enumerate(pep):
            seq[idx] = res
            if res in static_mods:
                if var_mods is None:
                    var_mods = []

                var_mods.insert(0, static_mods[res])

            mods[idx + 1] = var_mods

        return cls("".join(seq), mods, charge)

    @classmethod
    def from_massivekb(
        cls,
        sequence: str,
        charge: int | None = None,
    ) -> Peptide:
        """Create a Peptide from MassIVE-KB annotations.

        MassIVE-KB includes N-term carbamylation, NH3-loss, acetylation,
        as well as M oxidation, and deamidation of N and Q, in a
        manner that does not comply with the ProForma standard.

        Parameters
        ----------
        sequence : str
            The peptide sequence from MassIVE-KB
        charge : int, optional
            The charge state of the peptide.

        Returns
        -------
        Peptide
            The parsed MassIVE peptide after conversion to a ProForma
            format.
        """
        sequence = cls.massivekb_to_proforma(sequence, charge)
        return cls.from_proforma(sequence)

    @classmethod
    def massivekb_to_proforma(
        cls, sequence: str, charge: int | None = None
    ) -> str:
        """Convert a MassIVE-KB peptide sequence to ProForma.

        MassIVE-KB includes N-term carbamylation, NH3-loss, acetylation,
        as well as M oxidation, and deamidation of N and Q, in a
        manner that does not comply with ProForma.

        Parameters
        ----------
        sequence : str
            The peptide sequence from MassIVE-KB
        charge : int, optional
            The charge state of the peptide.

        Returns
        -------
        str
            The parsed MassIVE peptide after conversion to a ProForma
            format.
        """
        sequence = "".join(
            [
                MSKB_TO_UNIMOD.get(aa, aa)
                for aa in re.split(r"(?<=.)(?=[A-Z])", sequence)
            ]
        )
        if charge is not None:
            sequence += f"/{charge}"

        return sequence


@dataclass
class PeptideIons:
    """The ions generated by a peptide.

    Parameters
    ----------
    tokens : list[str]
        The string tokens that comprise the peptide sequence.
    precursor : float
        The monoisotopic m/z of the precursor ion.
    fragments : torch.Tensor[float]
        The generated fragment ions originated from the peptide.
    """

    tokens: list[str]
    precursor: float
    fragments: torch.Tensor[float]

    @property
    def sequence(self) -> str:
        """The peptide sequence."""
        return "".join(self.tokens)

    @property
    def b_ions(self) -> torch.Tensor[float]:
        """The b ion series."""
        return self.fragments[0, :, :]

    @property
    def y_ions(self) -> torch.Tensor[float]:
        """The y ion series."""
        return self.fragments[1, :, :]


@dataclass
class Molecule:
    """A representation of a molecule.

    Parameters
    ----------
    smiles : str
        A SMILES string defining the molecule.
    charge : int, optional
        The charge of the molecule.
    """

    smiles: str
    charge: int | None = None

    def __post_init__(self) -> None:
        """Validate parameters."""
        self._mol = Chem.MolFromSmiles(self.smiles)
        if self._mol is None:
            raise ValueError("The SMILES was invalid.")

        # Canonnicalize the SMILES string:
        self.smiles = Chem.MolToSmiles(self._mol)

    def show(self, **kwargs: dict) -> PngImageFile:
        """Show the molecule in 2D.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments passed to ``rdkit.Chem.Draw.MolToImage``
        """
        return Draw.MolToImage(self._mol, **kwargs)

    def to_selfies(self) -> str:
        """Convert SMILES to a SELFIES representaion."""
        return sf.encoder(self.smiles)

    @classmethod
    def from_selfies(
        cls,
        selfies: str,
        charge: int | None = None,
    ) -> Molecule:
        """Create a molecule from a SELFIES string.

        Parameters
        ----------
        selfies : str
            The SELFIES string defining the molecule.
        charge : int, optional
            The charge of the molecule.

        Returns
        -------
        Molecule
            The parsed Molecule.
        """
        return cls(sf.decoder(selfies), charge)


class MassSpectrum(MsmsSpectrum):
    """A mass spectrum.

    Parameters
    ----------
    filename: str
        The file from which the spectrum originated.
    scan_id: str
        The Hupo PSI standard scan identifier.
    mz: array of shape (n_peaks,)
        The m/z values.
    intensity: array of shape (n_peaks, )
        The intensity values.
    retention_time: float, optional
        The measured retention time.
    ion_mobility: float, optional
        The measured ion mobility.
    precursor_mz: float, optional
        The precursor ion m/z, if applicable.
    precursor_charge: int, optional
        The precursor charge, if applicable.
    label: str, optional
        A label for the mass spectrum. This is typically an
        annotation, such as the generating peptide sequence,
        but is distinct from spectrum_utils' annotation.
    """

    def __init__(
        self,
        filename: str,
        scan_id: str,
        mz: ArrayLike,
        intensity: ArrayLike,
        rt: ArrayLike | None = None,
        level: ArrayLike | None = None,
        fragment_label: ArrayLike | None = None,
        retention_time: float | None = None,
        ion_mobility: float | None = None,
        precursor_mz: float | None = None,
        precursor_charge: int | None = None,
        label: str | None = None,
        global_rt: float | None = None,
    ) -> None:
        """Initialize a MassSpectrum."""
        self.filename = filename
        self.scan_id = scan_id
        self.label = label

        # Not currently supported by spectrum_utils:
        self.ion_mobility = ion_mobility

        # spectrum_utils requires a precursor. Will remove after
        # I make a PR:
        if precursor_mz is None:
            precursor_mz = np.nan

        if precursor_charge is None:
            precursor_charge = np.nan

        super().__init__(
            identifier=str(scan_id),
            precursor_mz=precursor_mz,
            precursor_charge=precursor_charge,
            mz=mz,
            intensity=intensity,
            retention_time=retention_time,
        )

        self.rt = rt 
        self.level = level
        self.global_rt = global_rt
        self.fragment_label = fragment_label


    @property
    def usi(self) -> str:
        """The Universal Spectrum Identifier."""
        index_type, scan = self.scan_id.split("=")
        fname = re.sub(
            ".(gz|tar|tar.gz|zip|bz2|tar.bz2)$",
            "",
            self.filename,
            flags=re.IGNORECASE,
        )
        fname = re.sub(
            ".(raw|mzml|mzxml|mgf|d|wiff)$",
            "",
            fname,
            flags=re.IGNORECASE,
        )
        return ":".join([fname, index_type, scan])

    @property
    def precursor_mass(self) -> float:
        """The monoisotopic mass."""
        return (self.precursor_mz - PROTON) * self.precursor_charge

    def to_tensor(self) -> torch.tensor:
        """Combine the m/z and intensity arrays into a single tensor.

        Returns
        -------
        torch.tensor of shape (n_peaks, 2)
            The mass spectrum information.
        """
        return torch.tensor(np.vstack([self.mz, self.intensity, self.rt, self.level]).T)



"""Initialize the depthcharge package."""
from . import (
    data,
    encoders,
    feedforward,
    tokenizers,
    transformers,
)
from .primitives import (
    MassSpectrum,
    Molecule,
    Peptide,
    PeptideIons,
)
from .version import _get_version

__version__ = _get_version()



from .depthcharge.data.spectrum_datasets import AnnotatedSpectrumDataset
from .depthcharge.data.preprocessing import scale_to_unit_norm, scale_intensity
from .depthcharge.tokenizers import PeptideTokenizer
import torch
import numpy as np
import pytorch_lightning as pl
import os
import sys
import argparse
from lightning.pytorch import loggers as pl_loggers
from .utils import *
from .model import AugmentedSpec2Pep
from .augment import *
from datetime import datetime
import warnings
import json
warnings.filterwarnings('ignore')


def sequence():
  parser=argparse.ArgumentParser()
  parser.add_argument("spectrum_file")
  parser.add_argument("model", type= str, help="A path to a trained Cascadia model checkpoint.")
  parser.add_argument("-o", "--outfile", type= str, default= 'cascadia_results', help="Output file for inference.")
  parser.add_argument("-t", "--score_threshold", type= float, default= 0.8, help="Score threshold for Cascadia predictions.")
  parser.add_argument("-b", "--batch_size", type= int, default= 32, help="Number of spectra to include in a batch.")
  parser.add_argument("-w", "--width", type= int, default= 2, help="Number of adjacent scans to use when constructing each augmented spectrum.")
  parser.add_argument("-c", "--max_charge", type= int, default= 4, help="Maximum precursor charge state to consider")
  parser.add_argument("-p", "--modifications", type= str, default= 'mskb', help="A path to a json file containing a list of the PTMs to consider.")
  
  args = parser.parse_args(args=sys.argv[2:])

  spectrum_file = args.spectrum_file
  model_ckpt_path = args.model
  results_file = args.outfile
  batch_size = args.batch_size
  score_threshold = args.score_threshold
  augmentation_width = args.width
  max_charge = args.max_charge
  mods = args.modifications
  
  temp_path = os.getcwd() + '/cascadia_' +  datetime.now().strftime("%m-%d-%H:%M:%S")
  os.mkdir(temp_path)
  train_index_filename = temp_path + "/index.hdf5"

  print("Augmenting spectra from:", spectrum_file)
  asf_file, isolation_window_size, cycle_time = augment_spectra(spectrum_file, temp_path, max_charge=max_charge)
  
  if mods == 'mskb':
    tokenizer = PeptideTokenizer.from_massivekb(reverse=False, replace_isoleucine_with_leucine=True)
  else:
    with open('ptms.json', 'r') as f:
      proforma = json.load(f)
    tokenizer = PeptideTokenizer.from_proforma(proforma, reverse=False, replace_isoleucine_with_leucine=True)
  
  train_dataset = AnnotatedSpectrumDataset(tokenizer, asf_file, index_path=train_index_filename, preprocessing_fn=[scale_intensity(scaling="root"), scale_to_unit_norm])
  train_loader = train_dataset.loader(batch_size=batch_size, num_workers=4, pin_memory=True)

  if os.path.exists(asf_file):
      os.remove(asf_file)

  model = AugmentedSpec2Pep.load_from_checkpoint(
      model_ckpt_path,
      d_model = 512,
      n_layers = 9,
      n_head = 8,
      dim_feedforward = 1024,
      dropout = 0,
      rt_width = 2,
      tokenizer=tokenizer,
      max_charge=10,
  )

  print("Running inference on augmented spectra from:", spectrum_file)
  if torch.cuda.is_available():
    device = 'gpu'
    print('GPU found')
  else:
    device = 'cpu'
    print(f'No GPU found - running inference on cpu')

  trainer = pl.Trainer(max_epochs=50, log_every_n_steps=1, accelerator=device, devices = 1)
  preds = trainer.predict(model, dataloaders=train_loader)

  print("Writing results to:", results_file + '.ssl')

  write_results(preds, results_file, spectrum_file, isolation_window_size, score_threshold, augmentation_width*cycle_time)
  
  os.remove(train_index_filename)
  os.rmdir(temp_path)
  
  return parser

def train():
  print("train", sys.argv)
  parser=argparse.ArgumentParser()
  parser.add_argument("train_spectrum_file")
  parser.add_argument("val_spectrum_file")
  parser.add_argument("-m", "--model", type= str, required=False, help="A path to a Cascadia model checkpoint to fine tune.")
  parser.add_argument("-b", "--batch_size", type= int, default= 32, help="Number of spectra to include in a batch.")
  parser.add_argument("-w", "--width", type= int, default= 2, help="Number of adjacent scans to use when constructing each augmented spectrum.")
  parser.add_argument("-c", "--max_charge", type= int, default= 4, help="Maximum precursor charge state to consider.")
  parser.add_argument("-e", "--max_epochs", type= int, default= 10, help="Maximum number of training epochs.")
  parser.add_argument("-lr", "--learning_rate", type= float, default= 1e-5, help="Learning rate for model training.")
  parser.add_argument("-p", "--modifications", type= str, default= 'mskb', help="A path to a json file containing a list of the PTMs to consider.")
  
  args = parser.parse_args(args=sys.argv[2:])

  train_spectrum_file = args.train_spectrum_file
  val_spectrum_file = args.val_spectrum_file
  model_ckpt_path = args.model
  batch_size = args.batch_size
  augmentation_width = args.width
  max_charge = args.max_charge
  max_epochs = args.max_epochs
  lr = args.learning_rate
  mods = args.modifications
  
  if torch.cuda.is_available():
    device = 'gpu'
    print('GPU found')
  else:
    device = 'cpu'
    print("No GPU Available - training on CPU will be extremely slow!")

  print("Training on spectra from:", train_spectrum_file)
  print("Validating on spectra from:", val_spectrum_file)
  
  ckpt_path = os.getcwd() + '/checkpoint_' +  datetime.now().strftime("%m-%d-%H:%M:%S")
  os.mkdir(ckpt_path)
  train_index_filename = ckpt_path + "/train_gpu.hdf5"
  val_index_filename = ckpt_path + "/val_gpu.hdf5"

  if os.path.exists(train_index_filename):
    os.remove(train_index_filename)
  if os.path.exists(val_index_filename):
    os.remove(val_index_filename)

  if mods == 'mskb':
    tokenizer = PeptideTokenizer.from_massivekb(reverse=False, replace_isoleucine_with_leucine=True)
  else:
    with open(mods, 'r') as f:
      proforma = json.load(f)
    tokenizer = PeptideTokenizer.from_proforma(proforma, reverse=False, replace_isoleucine_with_leucine=True)
  
  if '.hdf5' in train_spectrum_file:
    train_dataset = AnnotatedSpectrumDataset(tokenizer, index_path=train_spectrum_file, preprocessing_fn=[scale_intensity(scaling="root"), scale_to_unit_norm])
    val_dataset = AnnotatedSpectrumDataset(tokenizer, index_path=val_spectrum_file, preprocessing_fn=[scale_intensity(scaling="root"), scale_to_unit_norm])
  else:
    train_dataset = AnnotatedSpectrumDataset(tokenizer, train_spectrum_file, index_path=train_index_filename, preprocessing_fn=[scale_intensity(scaling="root"), scale_to_unit_norm]) 
    val_dataset = AnnotatedSpectrumDataset(tokenizer, val_spectrum_file, index_path=val_index_filename, preprocessing_fn=[scale_intensity(scaling="root"), scale_to_unit_norm])
  
  train_loader = train_dataset.loader(batch_size=batch_size, num_workers=10, pin_memory=True, shuffle=True)
  val_loader = val_dataset.loader(batch_size=batch_size, num_workers=10, pin_memory=True)

  if model_ckpt_path is None:
    print("Training model from scratch") 
    model = AugmentedSpec2Pep(
        d_model = 512,
        n_layers = 9,
        n_head = 8,
        dim_feedforward = 1024,
        dropout = 0,
        rt_width = augmentation_width,
        tokenizer=tokenizer,
        max_charge=max_charge,
        lr=lr
    )
    
  else:
    print("Loading model from checkpoint:", model_ckpt_path) 
    model = AugmentedSpec2Pep.load_from_checkpoint(
      model_ckpt_path, 
      d_model = 512,
      n_layers = 9,
      n_head = 8,
      dim_feedforward = 1024,
      dropout = 0,
      rt_width = 2,
      tokenizer=tokenizer,
      max_charge=max_charge,
      lr=lr
    )

  tb_logger = pl_loggers.TensorBoardLogger(save_dir="logs/")
  ckpt_callback = pl.callbacks.ModelCheckpoint(
      dirpath=ckpt_path,
      filename="Cascadia-{epoch}-{step}",
      monitor="Val Pep. Acc.",
      mode='max',
      save_top_k=2
  )
  trainer = pl.Trainer(max_epochs=max_epochs, logger=tb_logger, log_every_n_steps=10000, val_check_interval = 10000, check_val_every_n_epoch=None, callbacks=[ckpt_callback], accelerator=device, devices=1)
  trainer.fit(model, train_loader, val_loader)
  
def main():
  parser=argparse.ArgumentParser()
  parser.add_argument("mode", choices=['sequence', 'train'], help='Which command to call')
  args = parser.parse_args(args=sys.argv[1:2])
  mode = args.mode
  if mode == 'sequence':
    sequence()
  elif mode == 'train':
    train()

if __name__ == '__main__':
    main()


"""Common utility functions."""
from typing import Any


def listify(obj: Any) -> list[Any]:  # noqa: ANN401
    """Turn an object into a list, but don't split strings."""
    try:
        assert not isinstance(obj, str)
        iter(obj)
    except (AssertionError, TypeError):
        obj = [obj]

    return list(obj)


def check_int(integer: int, name: str) -> int:
    """Verify that an object is an integer, or coercible to one.

    Parameters
    ----------
    integer : int
        The integer to check.
    name : str
        The name to print in the error message if it fails.

    Returns
    -------
    int
        The coerced integer.
    """
    if isinstance(integer, int):
        return integer

    # Else if it is a float:
    coerced = int(integer)
    if coerced != integer:
        raise ValueError(f"'{name}' must be an integer.")

    return coerced


def check_positive_int(integer: int, name: str) -> int:
    """Verify that an object is an integer and positive.

    Parameters
    ----------
    integer : int
        The integer to check.
    name : str
        The name to print in the error message if it fails.

    Returns
    -------
    int
        The coerced integer.
    """
    try:
        integer = check_int(integer, name)
        assert integer > 0
    except (ValueError, AssertionError):
        raise ValueError(f"'{name}' must be a positive integer.")

    return integer



from pyteomics import mzml, auxiliary, mass
import numpy as np
import os

def get_centers(mzml_file):
    f_to_mzrt_to_pep = {}
    max_mz = 0
    num_spectra = 0
    part = 0
    last_rt = 0
    cycle_time = None
    with mzml.read(mzml_file) as reader:
        for spec in reader:
            if spec['ms level'] == 1:
                cur_rt = 60 * spec['scanList']['scan'][0]['scan start time']
                cycle_time = cur_rt - last_rt
                last_rt = cur_rt
            if spec['ms level'] == 2:
                window = spec['precursorList']['precursor'][0]['isolationWindow']
                window_center = window['isolation window target m/z']
                lower_offset = window['isolation window lower offset']
                upper_offset = window['isolation window upper offset']
                window_size = upper_offset + lower_offset
                cur_rt = 60 * spec['scanList']['scan'][0]['scan start time']
                if num_spectra % 50000 == 0:
                    part += 1
                    f_to_mzrt_to_pep[part] = {}
                # if count > 1000:
                #     break
                num_spectra += 1
                key = (int(window_center/10), int(cur_rt/10))
                max_mz = max(max_mz,int(window_center/10))
                if key in f_to_mzrt_to_pep[part]:
                    f_to_mzrt_to_pep[part][key].append((window_center, cur_rt, 1))
                else:
                    f_to_mzrt_to_pep[part][key] = [(window_center, cur_rt, 1)]
    return f_to_mzrt_to_pep, max_mz, window_size, cycle_time

def extract_spectra(mzml_file, f_to_mzrt_to_pep, part, top_n, time_width, max_mz):
    prec_to_spec = {}
    with mzml.read(mzml_file) as reader:
        for spec in reader:
            cur_rt = 60 * spec['scanList']['scan'][0]['scan start time']
            if spec['ms level'] == 1:
                for scan_rt in range(int(cur_rt/10) - 1, int(cur_rt/10) + 1):
                    for scan_window in range(max_mz+1):
                        if (scan_window, scan_rt) in f_to_mzrt_to_pep[part]:
                            for mz, rt, charge in f_to_mzrt_to_pep[part][(scan_window, scan_rt)]:
                                if np.abs(rt - cur_rt) < time_width: 
                                    mzs = spec['m/z array']
                                    intensities = spec['intensity array']

                                    sorted_intensity_idxs = np.argsort(intensities)[-top_n:]
                                    intensities = intensities[sorted_intensity_idxs]
                                    mzs = mzs[sorted_intensity_idxs]

                                    sorted_mz_idxs = np.argsort(mzs)
                                    intensities = intensities[sorted_mz_idxs]
                                    mzs = mzs[sorted_mz_idxs]

                                    intensities = intensities ** .5
                                    intensities = intensities ** .5
                                    if len(intensities) > 0:
                                        intensities = intensities/np.max(intensities)

                                    if (mz, rt, charge) not in prec_to_spec:
                                        prec_to_spec[(mz, rt, charge)] = {}
                                    if 'ms1_scans' not in prec_to_spec[(mz, rt, charge)]:
                                        prec_to_spec[(mz, rt, charge)]['ms1_scans'] = []
                                        prec_to_spec[(mz, rt, charge)]['ms1_rts'] = []
                                    prec_to_spec[(mz, rt, charge)]['ms1_scans'].append([x for x in zip(mzs, intensities)])
                                    prec_to_spec[(mz, rt, charge)]['ms1_rts'].append(cur_rt - rt)
            elif spec['ms level'] == 2:
                window = spec['precursorList']['precursor'][0]['isolationWindow']
                window_center = window['isolation window target m/z']
                lower_offset = window['isolation window lower offset']
                upper_offset = window['isolation window upper offset']
            
                for scan_rt in range(int(cur_rt/10) - 1, int(cur_rt/10) + 1):
                    for scan_window in range(int((window_center - lower_offset)/10) - 1, int((window_center + upper_offset)/10) + 1):
                        if (scan_window, scan_rt) in f_to_mzrt_to_pep[part]:
                            for mz, rt, charge in f_to_mzrt_to_pep[part][(scan_window, scan_rt)]:
                                in_mz = mz > window_center - lower_offset and mz < window_center + upper_offset
                                rt_diff = np.abs(rt - cur_rt)
                                if in_mz and rt_diff < time_width:
                                    mzs = spec['m/z array']
                                    intensities = spec['intensity array']

                                    sorted_intensity_idxs = np.argsort(intensities)[-top_n:]
                                    intensities = intensities[sorted_intensity_idxs]
                                    mzs = mzs[sorted_intensity_idxs]

                                    sorted_mz_idxs = np.argsort(mzs)
                                    intensities = intensities[sorted_mz_idxs]
                                    mzs = mzs[sorted_mz_idxs]

                                    intensities = intensities ** .5
                                    if len(intensities) > 0:
                                        intensities = intensities/np.max(intensities)
                      
                                    if (mz, rt, charge) not in prec_to_spec:
                                        prec_to_spec[(mz, rt, charge)] = {}
                                    if 'scans' not in prec_to_spec[(mz, rt, charge)]:
                                        prec_to_spec[(mz, rt, charge)]['scans'] = []
                                        prec_to_spec[(mz, rt, charge)]['rts'] = []
                                        prec_to_spec[(mz, rt, charge)]['window_width'] = max(lower_offset, upper_offset) 
                                    prec_to_spec[(mz, rt, charge)]['scans'].append([x for x in zip(mzs, intensities)])
                                    prec_to_spec[(mz, rt, charge)]['rts'].append(cur_rt - rt)
    return prec_to_spec

def write_asf(outfile, prec_to_spec, scan_width, max_pep_length, max_charge):
    out = open(outfile, 'a')
    skipped = 0
    count = 0
    for key, value in prec_to_spec.items():
        if 'ms1_scans' not in value:
            skipped += 1
            continue
        prec, rt, charge = key
        scans = np.array(value['scans'], dtype=object)
        rts = np.array(value['rts'])
        ms1_scans = np.array(value['ms1_scans'], dtype=object)
        ms1_rts = np.array(value['ms1_rts'])
        window_width = value['window_width']

        abs_rts = [np.abs(x) for x in rts]
        sorted_rt_idxs = np.argsort(abs_rts)[:scan_width]
        rts = rts[sorted_rt_idxs]
        scans = scans[sorted_rt_idxs]

        abs_ms1_rts = [np.abs(x) for x in ms1_rts]
        sorted_ms1_rt_idxs = np.argsort(abs_ms1_rts)[:scan_width]
        ms1_rts = ms1_rts[sorted_ms1_rt_idxs]
        ms1_scans = ms1_scans[sorted_ms1_rt_idxs]
        
        for charge in range(2,max_charge+1):
            count += 1

            out.write("BEGIN IONS\n")
            out.write(f"TITLE={str(count)}\n")
            out.write(f"PEPMASS={prec}\n")
            out.write(f"CHARGE={charge}\n")
            out.write(f"SCAN={count}\n")
            out.write(f"RT={rt}\n")
            out.write(f"SEQ={'K'*max_pep_length}\n")
            
            for scan, cur_rt in zip(scans,rts):
                for mz, intensity in scan:
                    out.write(f"{mz}\t{intensity}\t{cur_rt}\t2\n")
            
            for scan, cur_rt in zip(ms1_scans, ms1_rts):
                for mz, intensity in scan:
                    if np.abs(mz - prec) < window_width + 1:
                        out.write(f"{mz}\t{intensity}\t{cur_rt}\t1\n")

            out.write("END IONS\n")   
    out.close()

def augment_spectra(mzml_file, temp_path, top_n = 150, scan_width = 1, max_pep_length = 30, max_charge = 3):
    outfile = temp_path + "/temp.asf"
    f_to_mzrt_to_pep, max_mz, window_size, cycle_time = get_centers(mzml_file)
    for part in f_to_mzrt_to_pep.keys():
        prec_to_spec = extract_spectra(mzml_file, f_to_mzrt_to_pep, part, top_n, (scan_width + 1) * cycle_time, max_mz)
        print(f'Generating {len(prec_to_spec) * max_charge} augmented spectra')
        write_asf(outfile, prec_to_spec, scan_width, max_pep_length, max_charge)
    return outfile, window_size, cycle_time


"""Get the version information."""
from importlib.metadata import version, PackageNotFoundError


def _get_version() -> str:
    """Return the version information for deptcharge."""
    try:
        return version("depthcharge-ms")
    except PackageNotFoundError:
        return None



"""Preprocessing functions for mass spectra.

These functions can be used with datasets (`SpectrumDataset` and
`AnnotatedSpectrumDataset`).

One or more preprocessing function can be applied to each mass spectrum during
dataset preparation using the `preprocessing_fn` parameter of these classes. To
apply preprocessing steps sequentially, pass a list of functions to this
argument.

We can also define custom preprocessing functions. All preprocessing functions
must accept a ``MassSpectrum`` as their only argument and return a
``MassSpectrum``. If a ``Mass Spectrum`` is invalid, the function should raise
a ``ValueError``.

### Examples

Remove the peaks around the precursor m/z then square root transform
intensities and scale to unit norm:
```Python
SpectrumDataset(
    ...,
    preprocessing_fn=[
        preprocessing.remove_precursor_peak,
        preprocessing.sqrt_and_norm,
    ],
)
```

Apply a custom function:
```Python
def my_func(spec: MassSpectrum) -> MassSpectrum:
    spec._inner._intensity = np.log(spectrum.intensity)

SpectrumDataset(
    ...,
    preprocessing_fn=my_func,
)
```

"""
from collections.abc import Callable
from functools import wraps

import numpy as np

from ..primitives import MassSpectrum


def scale_to_unit_norm(spectrum: MassSpectrum) -> MassSpectrum:
    """Scale intensities to unit norm."""
    spectrum._inner._intensity = (
        spectrum.intensity / np.sqrt(spectrum.intensity**2).sum()
    )
    return spectrum


def _spectrum_utils_fn(func: str) -> Callable:
    """Wrap spectrum_utils.spectrum.MmsmsSpectrum preprocessing methods."""

    @wraps(getattr(MassSpectrum, func))
    def wrapper(
        *args: tuple,
        **kwargs: dict,
    ) -> Callable:
        """Wrapper for spectrum_utils MsmsSpectrum methods.

        Parameters
        ----------
        func: Callable
            The preprocessing function. This should exactly match a
            MsmsSpectrum method.
        *args: tuple
            Arguments that are passed to the MsmsSpectrum method.
        **kwargs : dict
            Keyword arguments that are passed to the MsmsSpectrum method.

        Returns
        -------
        Callable
            A valid deptcharge preprocessing function.
        """

        @wraps(wrapper)
        def preprocess(spec: MassSpectrum) -> MassSpectrum:
            """The wrapped preprocessing function.

            Parameters
            ----------
            spec : MassSpectrum
                The mass spectrum to preprocess

            Returns
            -------
            MassSpectrum
                The processed mass spectrum.
            """
            # Call the spectrum_utils method:
            getattr(spec, func)(*args, **kwargs)
            return spec

        return preprocess

    return wrapper


def _add_spectrum_utils_methods() -> None:
    """Update this module with the spectrum_utils MsmsSpectrum methods."""
    sus_methods = [
        "filter_intensity",
        "remove_precursor_peak",
        "round",
        "scale_intensity",
        "set_mz_range",
    ]

    for method in sus_methods:
        globals()[method] = _spectrum_utils_fn(method)


_add_spectrum_utils_methods()



from importlib.metadata import version

# Configuration file for the Sphinx documentation builder.

# -- Project information

project = 'Cascadia'
copyright = '2024, Justin Sanders'
author = 'Justin Sanders'
# release = version("cascadia")
# version = ".".join(release.split(".")[:2])
version = '0.0.1'

# -- General configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'myst_parser',
    'sphinxarg.ext',
]

# The format for each file suffix:
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

html_theme = 'sphinx_rtd_theme'

# -- Options for EPUB output
epub_show_urls = 'footnote'



"""Mass spectrometry data parsers."""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from os import PathLike
from pathlib import Path

import numpy as np
from pyteomics.mgf import MGF
from pyteomics.mzml import MzML
from pyteomics.mzxml import MzXML
from tqdm.auto import tqdm
from pyteomics import mass


from .. import utils
from ..primitives import MassSpectrum

LOGGER = logging.getLogger(__name__)


class BaseParser(ABC):
    """A base parser class to inherit from.

    Parameters
    ----------
    ms_data_file : PathLike
        The peak file to parse.
    ms_level : int
        The MS level of the spectra to parse.
    preprocessing_fn : Callable or Iterable[Callable], optional
        The function(s) used to preprocess the mass spectra.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    id_type : str, optional
        The Hupo-PSI prefix for the spectrum identifier.
    """

    def __init__(
        self,
        ms_data_file: PathLike,
        ms_level: int,
        preprocessing_fn: Callable | Iterable[Callable] | None = None,
        valid_charge: Iterable[int] | None = None,
        id_type: str = "scan",
    ) -> None:
        """Initialize the BaseParser."""
        self.path = Path(ms_data_file)
        self.ms_level = ms_level
        if preprocessing_fn is None:
            self.preprocessing_fn = []
        else:
            self.preprocessing_fn = utils.listify(preprocessing_fn)

        self.valid_charge = None if valid_charge is None else set(valid_charge)
        self.id_type = id_type
        self.offset = None
        self.precursor_mz = []
        self.precursor_charge = []
        self.scan_id = []
        self.mz_arrays = []
        self.intensity_arrays = []
        self.annotations = None

    @abstractmethod
    def open(self) -> Iterable:
        """Open the file as an iterable."""

    @abstractmethod
    def parse_spectrum(self, spectrum: dict) -> MassSpectrum | None:
        """Parse a single spectrum.

        Parameters
        ----------
        spectrum : dict
            The dictionary defining the spectrum in a given format.

        Returns
        -------
        MassSpectrum or None
            The parsed mass spectrum or None if it is skipped.
        """

    def read(self) -> BaseParser:
        """Read the ms data file.

        Returns
        -------
        Self
        """
        n_skipped = 0
        with self.open() as spectra:
            for spectrum in tqdm(spectra, desc=str(self.path), unit="spectra"):
                try:
                    spectrum = self.parse_spectrum(spectrum)
                    if spectrum is None:
                        continue

                    if self.preprocessing_fn is not None:
                        for processor in self.preprocessing_fn:
                            spectrum = processor(spectrum)

                    self.mz_arrays.append(spectrum.mz)
                    self.intensity_arrays.append(spectrum.intensity)
                    self.precursor_mz.append(spectrum.precursor_mz)
                    self.precursor_charge.append(spectrum.precursor_charge)
                    self.scan_id.append(_parse_scan_id(spectrum.scan_id))
                    if self.annotations is not None:
                        self.annotations.append(spectrum.label)
                except (IndexError, KeyError, ValueError):
                    n_skipped += 1

        if n_skipped:
            LOGGER.warning(
                "Skipped %d spectra with invalid precursor info", n_skipped
            )

        self.precursor_mz = np.array(self.precursor_mz, dtype=np.float64)
        self.precursor_charge = np.array(
            self.precursor_charge,
            dtype=np.uint8,
        )

        self.scan_id = np.array(self.scan_id)

        # Build the index
        sizes = np.array([0] + [s.shape[0] for s in self.mz_arrays])
        self.offset = sizes[:-1].cumsum()
        self.mz_arrays = np.concatenate(self.mz_arrays).astype(np.float64)
        self.intensity_arrays = np.concatenate(self.intensity_arrays).astype(
            np.float32
        )
        return self

    @property
    def n_spectra(self) -> int:
        """The number of spectra."""
        return self.offset.shape[0]

    @property
    def n_peaks(self) -> int:
        """The number of peaks in the file."""
        return self.mz_arrays.shape[0]


class MzmlParser(BaseParser):
    """Parse mass spectra from an mzML file.

    Parameters
    ----------
    ms_data_file : PathLike
        The mzML file to parse.
    ms_level : int
        The MS level of the spectra to parse.
    preprocessing_fn : Callable or Iterable[Callable], optional
        The function(s) used to preprocess the mass spectra.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    """

    def open(self) -> Iterable[dict]:
        """Open the mzML file for reading."""
        return MzML(str(self.path))

    def parse_spectrum(self, spectrum: dict) -> MassSpectrum | None:
        """Parse a single spectrum.

        Parameters
        ----------
        spectrum : dict
            The dictionary defining the spectrum in mzML format.

        Returns
        -------
        MassSpectrum or None
            The parsed mass spectrum or None if not at the correct MS level.
        """
        if spectrum["ms level"] != self.ms_level:
            return None

        if self.ms_level > 1:
            precursor = spectrum["precursorList"]["precursor"][0]
            precursor_ion = precursor["selectedIonList"]["selectedIon"][0]
            precursor_mz = float(precursor_ion["selected ion m/z"])
            if "charge state" in precursor_ion:
                precursor_charge = int(precursor_ion["charge state"])
            elif "possible charge state" in precursor_ion:
                precursor_charge = int(precursor_ion["possible charge state"])
            else:
                precursor_charge = 0
        else:
            precursor_mz, precursor_charge = None, 0

        if self.valid_charge is None or precursor_charge in self.valid_charge:
            return MassSpectrum(
                filename=str(self.path),
                scan_id=spectrum["id"],
                mz=spectrum["m/z array"],
                intensity=spectrum["intensity array"],
                precursor_mz=precursor_mz,
                precursor_charge=precursor_charge,
            )

        raise ValueError("Invalid precursor charge")


class MzxmlParser(BaseParser):
    """Parse mass spectra from an mzXML file.

    Parameters
    ----------
    ms_data_file : PathLike
        The mzXML file to parse.
    ms_level : int
        The MS level of the spectra to parse.
    preprocessing_fn : Callable or Iterable[Callable], optional
        The function(s) used to preprocess the mass spectra.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    """

    def open(self) -> Iterable[dict]:
        """Open the mzXML file for reading."""
        return MzXML(str(self.path))

    def parse_spectrum(self, spectrum: dict) -> MassSpectrum | None:
        """Parse a single spectrum.

        Parameters
        ----------
        spectrum : dict
            The dictionary defining the spectrum in mzXML format.

        Returns
        -------
        MassSpectrum
            The parsed mass spectrum.
        """
        if spectrum["msLevel"] != self.ms_level:
            return None

        if self.ms_level > 1:
            precursor = spectrum["precursorMz"][0]
            precursor_mz = float(precursor["precursorMz"])
            precursor_charge = int(precursor.get("precursorCharge", 0))
        else:
            precursor_mz, precursor_charge = None, 0

        if self.valid_charge is None or precursor_charge in self.valid_charge:
            return MassSpectrum(
                filename=str(self.path),
                scan_id=spectrum["id"],
                mz=spectrum["m/z array"],
                intensity=spectrum["intensity array"],
                precursor_mz=precursor_mz,
                precursor_charge=precursor_charge,
            )

        raise ValueError("Invalid precursor charge")


class MgfParser(BaseParser):
    """Parse mass spectra from an MGF file.

    Parameters
    ----------
    ms_data_file : PathLike
        The MGF file to parse.
    ms_level : int
        The MS level of the spectra to parse.
    preprocessing_fn : Callable or Iterable[Callable], optional
        The function(s) used to preprocess the mass spectra.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    annotations : bool
        Include peptide annotations.
    """

    def __init__(
        self,
        ms_data_file: PathLike,
        ms_level: int = 2,
        preprocessing_fn: Callable | Iterable[Callable] | None = None,
        valid_charge: Iterable[int] | None = None,
        annotations: bool = False,
    ) -> None:
        """Initialize the MgfParser."""
        super().__init__(
            ms_data_file,
            ms_level=ms_level,
            preprocessing_fn=preprocessing_fn,
            valid_charge=valid_charge,
            id_type="index",
        )
        if annotations:
            self.annotations = []

        self._counter = -1

    def open(self) -> Iterable[dict]:
        """Open the MGF file for reading."""
        return MGF(str(self.path))

    def parse_spectrum(self, spectrum: dict) -> MassSpectrum:
        """Parse a single spectrum.

        Parameters
        ----------
        spectrum : dict
            The dictionary defining the spectrum in MGF format.
        """
        self._counter += 1

        if self.ms_level > 1:
            precursor_mz = float(spectrum["params"]["pepmass"][0])
            precursor_charge = int(spectrum["params"].get("charge", [0])[0])
        else:
            precursor_mz, precursor_charge = None, 0

        if self.annotations is not None:
            label = spectrum["params"].get("seq")
        else:
            label = None

        if self.valid_charge is None or precursor_charge in self.valid_charge:
            return MassSpectrum(
                filename=str(self.path),
                scan_id=self._counter,
                mz=spectrum["m/z array"],
                intensity=spectrum["intensity array"],
                precursor_mz=precursor_mz,
                precursor_charge=precursor_charge,
                label=label,
            )

        raise ValueError("Invalid precursor charge")
    
class AsfParser(BaseParser):
    """Parse augmented mass spectra from an asf file.

    Parameters
    ----------
    ms_data_file : PathLike
        The asf file to parse.
    ms_level : int
        The MS level of the spectra to parse.
    preprocessing_fn : Callable or Iterable[Callable], optional
        The function(s) used to preprocess the mass spectra.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    annotations : bool
        Include peptide annotations.
    """

    def __init__(
        self,
        ms_data_file: PathLike,
        ms_level: int = 2,
        preprocessing_fn: Callable | Iterable[Callable] | None = None,
        valid_charge: Iterable[int] | None = None,
        annotations: bool = False,
    ) -> None:
        """Initialize the MgfParser."""
        super().__init__(
            ms_data_file,
            ms_level=ms_level,
            preprocessing_fn=preprocessing_fn,
            valid_charge=valid_charge,
            id_type="index",
        )
        if annotations:
            self.annotations = []

        self.rt_arrays = []
        self.level_arrays = []
        self.fragment_arrays = []
        self.global_rts = []
        self._counter = -1
    
    def parse_spectrum(self, spectrum: dict) -> MassSpectrum:
        pass

    def open(self) -> Iterable[dict]:
        pass
    
    def read(self) -> BaseParser:
        """Read the ms data file.

        Returns
        -------
        Self
        """
        n_skipped = 0
        with open(self.path) as asf:
            for line in tqdm(asf, desc=str(self.path), unit="lines"):
                line = line.split()
                if line[0] == 'BEGIN':
                    mzs = []
                    intensities = []
                    rts = []
                    levels = []
                    precursor_mz = None
                    precursor_charge = 0
                    scan_id = 0
                    label = None
                    rt = None
                elif line[0][:7] == 'PEPMASS':
                    precursor_mz = float(line[0].split('=')[1])
                elif line[0][:6] == 'CHARGE':
                    precursor_charge = float(line[0][7])
                elif line[0][:4] == 'SCAN':
                    scan_id = float(line[0].split('=')[1])
                elif line[0][:3] == 'SEQ':
                    label = line[0].split('=')[1]
                elif line[0][:3] == 'RT=':
                    rt = float(line[0].split('=')[1])
                elif len(line) == 4 and '.' in line[0] and '.' in line[1]:
                    mzs.append(float(line[0]))
                    intensities.append(float(line[1]))
                    rts.append(float(line[2]))
                    levels.append(float(line[3]))
                elif line[0] == 'END':

                    seq = ''.join(filter(str.isalpha, label))
                    fragments = []
                    fragment_labels = []
                    #for i in range(1, len(seq)):
                        # fragments.append(mass.fast_mass(seq[:i], ion_type='b', charge=1) + 57.021 * seq.count('C'))
                        # fragments.append(mass.fast_mass(seq[i:], ion_type='y', charge=1) + 57.021 * seq.count('C'))
                        # fragments.append(mass.fast_mass(seq[:i], ion_type='b', charge=2) + 57.021/2 * seq.count('C'))
                        # fragments.append(mass.fast_mass(seq[i:], ion_type='y', charge=2) + 57.021/2 * seq.count('C'))

                    for peak in mzs:
                        peak_label = 0
                    #     for frag in fragments:
                    #         if np.abs(peak - frag) < (20 * frag / 1e6):
                    #             peak_label = 1
                    #             break
                        fragment_labels.append(peak_label)

                    spectrum = MassSpectrum(
                        filename=str(self.path),
                        scan_id=scan_id,
                        mz=mzs,
                        intensity=intensities,
                        rt=np.array(rts),
                        level=np.array(levels),
                        precursor_mz=precursor_mz,
                        precursor_charge=precursor_charge,
                        label=label,
                        fragment_label=np.array(fragment_labels)
                    )
                    
                    if self.preprocessing_fn is not None:
                        for processor in self.preprocessing_fn:
                            spectrum = processor(spectrum)

                    self.global_rts.append(rt)
                    self.mz_arrays.append(spectrum.mz)
                    self.intensity_arrays.append(spectrum.intensity)
                    self.rt_arrays.append(spectrum.rt)
                    self.level_arrays.append(spectrum.level)
                    self.fragment_arrays.append(spectrum.fragment_label)
                    self.precursor_mz.append(spectrum.precursor_mz)
                    self.precursor_charge.append(spectrum.precursor_charge)
                    self.scan_id.append(_parse_scan_id(spectrum.scan_id))
                    if self.annotations is not None:
                        self.annotations.append(spectrum.label)

        if n_skipped:
            LOGGER.warning(
                "Skipped %d spectra with invalid precursor info", n_skipped
            )

        self.precursor_mz = np.array(self.precursor_mz, dtype=np.float64)
        self.precursor_charge = np.array(
            self.precursor_charge,
            dtype=np.uint8,
        )

        self.scan_id = np.array(self.scan_id)
        self.global_rts = np.array(self.global_rts)

        # Build the index
        sizes = np.array([0] + [s.shape[0] for s in self.mz_arrays])
        self.offset = sizes[:-1].cumsum()
        self.mz_arrays = np.concatenate(self.mz_arrays).astype(np.float64)
        self.intensity_arrays = np.concatenate(self.intensity_arrays).astype(
            np.float32
        )
        self.rt_arrays = np.concatenate(self.rt_arrays).astype(np.float32)
        self.level_arrays = np.concatenate(self.level_arrays).astype(np.uint8)
        self.fragment_arrays = np.concatenate(self.fragment_arrays).astype(np.uint8)

        return self


def _parse_scan_id(scan_str: str | int) -> int:
    """Remove the string prefix from the scan ID.

    Adapted from:
    https://github.com/bittremieux/GLEAMS/blob/
    8831ad6b7a5fc391f8d3b79dec976b51a2279306/gleams/
    ms_io/mzml_io.py#L82-L85

    Parameters
    ----------
    scan_str : str
        The scan ID string.

    Returns
    -------
    int
        The scan ID number.
    """
    try:
        return int(scan_str)
    except ValueError:
        try:
            return int(scan_str[scan_str.find("scan=") + len("scan=") :])
        except ValueError:
            pass

    raise ValueError("Failed to parse scan number")



"""Tranformer models to handle mass spectra."""
from collections.abc import Callable

import torch

from ..encoders import PeakEncoder


class SpectrumTransformerEncoder(torch.nn.Module):
    """A Transformer encoder for input mass spectra.

    Parameters
    ----------
    d_model : int, optional
        The latent dimensionality to represent peaks in the mass spectrum.
    n_head : int, optional
        The number of attention heads in each layer. ``d_model`` must be
        divisible by ``nhead``.
    dim_feedforward : int, optional
        The dimensionality of the fully connected layers in the Transformer
        layers of the model.
    n_layers : int, optional
        The number of Transformer layers.
    dropout : float, optional
        The dropout probability for all layers.
    peak_encoder : PeakEncoder or bool, optional
        Sinusoidal encodings m/z and intensityvalues of each peak.
    """

    def __init__(
        self,
        d_model: int = 128,
        n_head: int = 8,
        dim_feedforward: int = 1024,
        n_layers: int = 1,
        dropout: float = 0,
        peak_encoder: PeakEncoder | Callable | bool = True,
    ) -> None:
        """Initialize a SpectrumEncoder."""
        super().__init__()

        self.latent_spectrum = torch.nn.Parameter(torch.randn(1, 1, d_model))

        if callable(peak_encoder):
            self.peak_encoder = peak_encoder
        elif peak_encoder:
            self.peak_encoder = PeakEncoder(d_model)
        else:
            self.peak_encoder = torch.nn.Linear(2, d_model)

        # The Transformer layers:
        layer = torch.nn.TransformerEncoderLayer(
            d_model=d_model,
            nhead=n_head,
            dim_feedforward=dim_feedforward,
            batch_first=True,
            dropout=dropout,
        )

        self.transformer_encoder = torch.nn.TransformerEncoder(
            layer,
            num_layers=n_layers,
        )

    def forward(
        self,
        spectra: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Embed a mass spectrum.

        Parameters
        ----------
        spectra : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra to embed. Axis 0 represents a mass spectrum, axis 1
            contains the peaks in the mass spectrum, and axis 2 is essentially
            a 2-tuple specifying the m/z-intensity pair for each peak. These
            should be zero-padded, such that all of the spectra in the batch
            are the same length.

        Returns
        -------
        latent : torch.Tensor of shape (n_spectra, n_peaks + 1, d_model)
            The latent representations for the spectrum and each of its
            peaks.
        mem_mask : torch.Tensor
            The memory mask specifying which elements were padding in X.
        """
        zeros = ~spectra.sum(dim=2).bool()
        mask = [
            torch.tensor([[False]] * spectra.shape[0]).type_as(zeros),
            zeros,
        ]
        mask = torch.cat(mask, dim=1)
        peaks = self.peak_encoder(spectra)

        # Add the spectrum representation to each input:
        latent_spectra = self.latent_spectrum.expand(peaks.shape[0], -1, -1)

        peaks = torch.cat([latent_spectra, peaks], dim=1)
        return self.transformer_encoder(peaks, src_key_padding_mask=mask), mask

    @property
    def device(self) -> torch.device:
        """The current device for the model."""
        return next(self.parameters()).device



"""A base Tokenizer class."""
from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence

import torch
from sortedcontainers import SortedDict, SortedSet
from torch import nn

from .. import utils


class Tokenizer(ABC):
    """An abstract base class for Depthcharge tokenizers.

    Parameters
    ----------
    tokens : Sequence[str]
        The tokens to consider.
    stop_token : str, optional
        The stop token to use.
    """

    def __init__(self, tokens: Sequence[str], stop_token: str = "$") -> None:
        """Initialize a tokenizer."""
        self.stop_token = stop_token

        tokens = SortedSet(tokens)
        if self.stop_token in tokens:
            raise ValueError(
                f"Stop token {stop_token} already exists in tokens.",
            )

        tokens.add(self.stop_token)
        self.index = SortedDict({k: i + 1 for i, k in enumerate(tokens)})
        self.reverse_index = [None] + list(tokens)
        self.stop_int = self.index[self.stop_token]

    def __len__(self) -> int:
        """The number of tokens."""
        return len(self.index)

    @abstractmethod
    def split(self, sequence: str) -> list[str]:
        """Split a sequence into the constituent string tokens."""

    def tokenize(
        self,
        sequences: Iterable[str],
        to_strings: bool = False,
        add_stop: bool = False,
    ) -> torch.Tensor | list[list[str]]:
        """Tokenize the input sequences.

        Parameters
        ----------
        sequences : Iterable[str]
            The sequences to tokenize.
        to_strings : bool, optional
            Return each as a list of token strings rather than a
            tensor. This is useful for debugging.
        add_stop : bool, optional
            Append the stop token tothe end of the sequence.

        Returns
        -------
        torch.Tensor of shape (n_sequences, max_length) or list[list[str]]
            Either a tensor containing the integer values for each
            token, padded with 0's, or the list of tokens comprising
            each sequence.
        """
        try:
            out = []
            for seq in utils.listify(sequences):
                tokens = self.split(seq)
                if add_stop and tokens[-1] != self.stop_token:
                    tokens.append(self.stop_token)

                if to_strings:
                    out.append(tokens)
                    continue

                out.append(torch.tensor([self.index[t] for t in tokens]))

            if to_strings:
                return out

            if isinstance(sequences, str):
                return out[0]

            return nn.utils.rnn.pad_sequence(out, batch_first=True)
        except KeyError as err:
            raise ValueError("Unrecognized token") from err

    def detokenize(
        self,
        tokens: torch.Tensor,
        join: bool = True,
        trim_stop_token: bool = True,
    ) -> list[str] | list[list[str]]:
        """Retreive sequences from tokens.

        Parameters
        ----------
        tokens : torch.Tensor of shape (n_sequences, max_length)
            The zero-padded tensor of integerized tokens to decode.
        join : bool, optional
            Join tokens into strings?
        trim_stop_token : bool, optional
            Remove the stop token from the end of a sequence.

        Returns
        -------
        list[str] or list[list[str]]
            The decoded sequences each as a string or list or strings.
        """
        decoded = []
        for row in tokens:
            seq = [
                self.reverse_index[i]
                for i in row
                if self.reverse_index[i] is not None
            ]

            if trim_stop_token and seq[-1] == self.stop_token:
                seq.pop(-1)

            if join:
                seq = "".join(seq)

            decoded.append(seq)

        return decoded



"""Tokenizers for peptides."""
from __future__ import annotations

import re
from collections.abc import Iterable

import numba as nb
import numpy as np
import torch
from pyteomics.proforma import GenericModification, MassModification

from .. import utils
from ..constants import H2O, PROTON
from ..primitives import MSKB_TO_UNIMOD, Peptide, PeptideIons
from .tokenizer import Tokenizer


class PeptideTokenizer(Tokenizer):
    """A tokenizer for ProForma peptide sequences.

    Parse and tokenize ProForma-compliant peptide sequences. Additionally,
    use this class to calculate fragment and precursor ion masses.

    Parameters
    ----------
    residues : dict[str, float], optional
        Residues and modifications to add to the vocabulary beyond the
        standard 20 amino acids.
    replace_isoleucine_with_leucine : bool
        Replace I with L residues, because they are isobaric and often
        indistinguishable by mass spectrometry.
    reverse : bool
        Reverse the sequence for tokenization, C-terminus to N-terminus.

    Attributes
    ----------
    residues : numba.typed.Dict[str, float]
        The residues and modifications and their associated masses.
        terminal modifcations are indicated by `-`.
    index : SortedDict{str, int}
        The mapping of residues and modifications to integer representations.
    reverse_index : list[None | str]
        The ordered residues and modifications where the list index is the
        integer representation for a token.
    stop_token : str
        The stop token.
    """

    residues = nb.typed.Dict.empty(
        nb.types.unicode_type,
        nb.types.float64,
    )
    residues.update(
        G=57.021463735,
        A=71.037113805,
        S=87.032028435,
        P=97.052763875,
        V=99.068413945,
        T=101.047678505,
        C=103.009184505,
        L=113.084064015,
        I=113.084064015,
        N=114.042927470,
        D=115.026943065,
        Q=128.058577540,
        K=128.094963050,
        E=129.042593135,
        M=131.040484645,
        H=137.058911875,
        F=147.068413945,
        R=156.101111050,
        Y=163.063328575,
        W=186.079312980,
    )

    # The peptide parsing function:
    _parse_peptide = Peptide.from_proforma

    def __init__(
        self,
        residues: dict[str, float] | None = None,
        replace_isoleucine_with_leucine: bool = False,
        reverse: bool = False,
    ) -> None:
        """Initialize a PeptideTokenizer."""
        self.replace_isoleucine_with_leucine = replace_isoleucine_with_leucine
        self.reverse = reverse
        self.residues = self.residues.copy()
        if residues is not None:
            self.residues.update(residues)

        if self.replace_isoleucine_with_leucine:
            del self.residues["I"]

        super().__init__(list(self.residues.keys()))

    def split(self, sequence: str) -> list[str]:
        """Split a ProForma peptide sequence.

        Parameters
        ----------
        sequence : str
            The peptide sequence.

        Returns
        -------
        list[str]
            The tokens that compprise the peptide sequence.
        """
        pep = self._parse_peptide(sequence)
        if self.replace_isoleucine_with_leucine:
            pep.sequence = pep.sequence.replace("I", "L")

        pep = pep.split()
        if self.reverse:
            pep.reverse()

        return pep

    def ions(  # noqa: C901
        self,
        sequences: Iterable[str],
        precursor_charges: Iterable[int] | str,
        max_fragment_charge: int | None = None,
    ) -> tuple[torch.Tensor[float], list[torch.Tensor[float]]]:
        """Calculate the m/z for the precursor and fragment ions.

        Currently depthcharge only support b and y ions.

        Parameters
        ----------
        sequences : Iterable[str],
            The peptide sequences.
        precursor_charges : Iterable[int] or None, optional
            The charge of each precursor ion. If ``None``, the charge state
            is expected to be found in the peptide strings.
        max_fragment_charge : int or None, optional
            The maximum charge for fragment ions. The default is to consider
            up to the ``max(precursor_charge - 1, 1)``.

        Returns
        -------
        list of PeptideIons
            The precursor and fragment ions generated by the peptide.
        """
        sequences = utils.listify(sequences)
        if max_fragment_charge is None:
            max_fragment_charge = np.inf

        if precursor_charges is None:
            precursor_charges = [None] * len(sequences)
        else:
            precursor_charges = utils.listify(precursor_charges)

        if len(sequences) != len(precursor_charges):
            raise ValueError(
                "The number of sequences and precursor charges did not match."
            )

        out = []
        for seq, charge in zip(sequences, precursor_charges):
            if isinstance(seq, str):
                if self.replace_isoleucine_with_leucine:
                    seq = seq.replace("I", "L")

                try:
                    pep = Peptide.from_proforma(seq)
                except ValueError:
                    pep = Peptide.from_massivekb(seq)

                tokens = pep.split()
                if charge is None:
                    charge = max(pep.charge - 1, 1)
            else:
                tokens = seq

            if charge is None:
                raise ValueError(
                    f"No charge was provided for {seq}",
                )

            try:
                prec = _calc_precursor_mass(
                    nb.typed.List(tokens),
                    charge,
                    self.residues,
                )
            except KeyError as err:
                raise ValueError(
                    f"Unrecognized token(s) in {''.join(tokens)}"
                ) from err

            frags = _calc_fragment_masses(
                nb.typed.List(tokens),
                min(charge, max_fragment_charge),
                self.residues,
            )

            ions = PeptideIons(
                tokens=tokens,
                precursor=prec,
                fragments=torch.tensor(frags),
            )

            out.append(ions)

        return out

    @classmethod
    def from_proforma(
        cls,
        sequences: Iterable[str],
        replace_isoleucine_with_leucine: bool = True,
        reverse: bool = True,
    ) -> PeptideTokenizer:
        """Create a tokenizer with the observed peptide modications.

        Modifications are parsed from ProForma 2.0-compliant peptide strings
        and added to the vocabulary.

        Parameters
        ----------
        sequences : Iterable[str]
            The peptides from which to parse modifications.
        replace_isoleucine_with_leucine : bool
            Replace I with L residues, because they are isobaric and often
            indistinguishable by mass spectrometry.
        reverse : bool
            Reverse the sequence for tokenization, C-terminus to N-terminus.

        Returns
        -------
        PeptideTokenizer
            A tokenizer for peptides with the observed modifications.
        """
        if isinstance(sequences, str):
            sequences = [sequences]

        # Parse modifications:
        new_res = cls.residues.copy()
        for peptide in sequences:
            parsed = Peptide.from_proforma(peptide).split()
            for token in parsed:
                if token in new_res.keys():
                    continue

                if token == "-":
                    continue

                match = re.search(r"(.*)\[(.*)\]", token)
                try:
                    res, mod = match.groups()
                    if res and res != "-":
                        res_mass = new_res[res]
                    else:
                        res_mass = 0
                except (AttributeError, KeyError) as err:
                    raise ValueError("Unrecognized token {token}.") from err

                try:
                    mod = MassModification(mod)
                except ValueError:
                    mod = GenericModification(mod)

                new_res[token] = res_mass + mod.mass

        return cls(new_res, replace_isoleucine_with_leucine, reverse)

    @staticmethod
    def from_massivekb(
        replace_isoleucine_with_leucine: bool = True,
        reverse: bool = True,
    ) -> MskbPeptideTokenizer:
        """Create a tokenizer with the observed peptide modications.

        Modifications are parsed from MassIVE-KB peptide strings
        and added to the vocabulary.

        Parameters
        ----------
        replace_isoleucine_with_leucine : bool
            Replace I with L residues, because they are isobaric and often
            indistinguishable by mass spectrometry.
        reverse : bool
            Reverse the sequence for tokenization, C-terminus to N-terminus.

        Returns
        -------
        MskbPeptideTokenizer
            A tokenizer for peptides with the observed modifications.
        """
        return MskbPeptideTokenizer.from_proforma(
            [f"{mod}A" for mod in MSKB_TO_UNIMOD.values()],
            replace_isoleucine_with_leucine,
            reverse,
        )


class MskbPeptideTokenizer(PeptideTokenizer):
    """A tokenizer for MassIVE-KB peptide sequences.

    MassIVE-KB includes N-term carbamylation, NH3-loss, acetylation, as well as
    M oxidation, and deamidation of N and Q, in a manner that does not comply
    with the ProForma standard.

    Parameters
    ----------
    replace_isoleucine_with_leucine : bool
        Replace I with L residues, because they are isobaric and often
        indistinguishable by mass spectrometry.
    reverse : bool
        Reverse the sequence for tokenization, C-terminus to N-terminus.

    Attributes
    ----------
    residues : SortedDict[str, float]
        The residues and modifications and their associated masses.
        terminal modifcations are indicated by `-`.
    index : SortedDicte{str, int}
        The mapping of residues and modifications to integer representations.
    reverse_index : list[None | str]
        The ordered residues and modifications where the list index is the
        integer representation for a token.
    stop_token : str
        The stop token.

    """

    _parse_peptide = Peptide.from_massivekb


@nb.njit
def _calc_precursor_mass(
    tokens: list[str],
    charge: int,
    masses: nb.typed.Dict,
) -> float:
    """Calculate the precursor mass of a peptide sequence.

    Parameters
    ----------
    tokens : list of str
        The tokenized peptide sequence.
    charge : int
        The charge state to consider. Use 'None' to get the neutral mass.
    masses : nb.typed.Dict
        The mass dictionary to use.

    Returns
    -------
    float
        The precurosr monoisotopic m/z.
    """
    mass = sum([masses[t] for t in tokens]) + H2O
    if charge is not None:
        mass = _mass2mz(mass, charge)

    return mass


@nb.njit
def _calc_fragment_masses(
    tokens: list[str],
    charge: int,
    masses: nb.typed.Dict,
) -> np.ndarray[float]:
    """Calculate the b and y ions for a peptide sequence.

    Parameters
    ----------
    tokens : list of str
        The tokenized peptide sequence.
    charge : int
        The charge state to consider. Use 'None' to get the neutral mass.
    masses : nb.typed.Dict
        The mass dictionary to use.

    Returns
    -------
    np.ndarray of shape (2, len(seq) - 1, charge)
        The m/z of the predicted b and y ions.
    """
    # Handle terminal mods:
    seq = np.empty(len(tokens))
    n_mod = False
    c_mod = False
    for idx, token in enumerate(tokens):
        if not idx and token.endswith("-"):
            n_mod = True

        if idx == (len(tokens) - 1) and token.startswith("-"):
            c_mod = True

        seq[idx] = masses[token]

    if n_mod:
        seq[1] += seq[0]
        seq = seq[1:]

    if c_mod:
        seq[-2] += seq[-1]
        seq = seq[:-1]

    # Calculate fragments:
    max_charge = min(charge, 2)
    n_ions = len(seq) - 1
    ions = np.empty((2, n_ions, max_charge))
    b_mass = 0
    y_mass = H2O
    for idx in range(n_ions):
        b_mass += seq[idx]
        y_mass += seq[-(idx + 1)]
        for cur_charge in range(1, max_charge + 1):
            z_idx = cur_charge - 1
            ions[0, idx, z_idx] = _mass2mz(b_mass, cur_charge)
            ions[1, idx, z_idx] = _mass2mz(y_mass, cur_charge)

    return ions


@nb.njit
def _mass2mz(mass: float, charge: int) -> float:
    """Calculate the m/z.

    Parameters
    ----------
    mass : float
        The neutral mass.
    charge : int
        The charge.

    Returns
    -------
    float
       The m/z
    """
    return (mass / charge) + PROTON



"""Parse mass spectra into an HDF5 file format."""
from __future__ import annotations

import hashlib
import logging
import uuid
from collections.abc import Callable, Iterable
from os import PathLike
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any
from pyteomics import mass

import dill
import h5py
import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset

from .. import utils
from ..primitives import MassSpectrum
from ..tokenizers import PeptideTokenizer
from . import preprocessing
from .parsers import MgfParser, MzmlParser, MzxmlParser, AsfParser

LOGGER = logging.getLogger(__name__)


class SpectrumDataset(Dataset):
    """Store and access a collection of mass spectra.

    Parses one or more MS data files, adding the mass spectra to an HDF5
    file that indexes the mass spectra from one or multiple files.
    This allows depthcharge to access random mass spectra from many different
    files quickly and without loading them all into memory.

    Parameters
    ----------
    ms_data_files : PathLike or list of PathLike, optional
        The mzML, mzXML, or MGF files to include in this collection. Files
        can be added later using the ``.add_file()`` method.
    ms_level : int, optional
        The level of tandem mass spectra to use.
    preprocessing_fn : Callable or Iterable[Callable], optional
        The function(s) used to preprocess the mass spectra. ``None``,
        the default, filters for the top 200 peaks above m/z 140,
        square root transforms the intensities and scales them to unit norm.
        See the preprocessing module for details and additional options.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    index_path : PathLike, optional.
        The name and path of the HDF5 file index. If the path does
        not contain the `.h5` or `.hdf5` extension, `.hdf5` will be added.
        If ``None``, a file will be created in a temporary directory.
    overwrite : bool, optional
        Overwrite previously indexed files? If ``False`` and new files are
        provided, they will be appended to the collection.

    Attributes
    ----------
    ms_files : list of str
    path : Path
    ms_level : int
    valid_charge : Optional[Iterable[int]]
    annotated : bool
    overwrite : bool
    n_spectra : int
    n_peaks : int
    """

    _annotated = False

    def __init__(
        self,
        ms_data_files: PathLike | Iterable[PathLike] = None,
        ms_level: int = 2,
        preprocessing_fn: Callable | Iterable[Callable] | None = None,
        valid_charge: Iterable[int] | None = None,
        index_path: PathLike | None = None,
        overwrite: bool = False,
    ) -> None:
        """Initialize a SpectrumIndex."""
        self._tmpdir = None
        if index_path is None:
            # Create a random temporary file:
            self._tmpdir = TemporaryDirectory()
            index_path = Path(self._tmpdir.name) / f"{uuid.uuid4()}.hdf5"

        index_path = Path(index_path)
        if index_path.suffix not in [".h5", ".hdf5"]:
            index_path = Path(index_path)
            index_path = Path(str(index_path) + ".hdf5")

        # Set attributes and check parameters:
        self._path = index_path
        self._ms_level = utils.check_positive_int(ms_level, "ms_level")
        self._valid_charge = valid_charge
        self._overwrite = bool(overwrite)
        self._handle = None
        self._file_offsets = np.array([0])
        self._file_map = {}
        self._locs = {}
        self._offsets = None

        if preprocessing_fn is not None:
            self._preprocessing_fn = utils.listify(preprocessing_fn)
        else:
            self._preprocessing_fn = [
                preprocessing.set_mz_range(min_mz=140),
                preprocessing.filter_intensity(max_num_peaks=200),
                preprocessing.scale_intensity(scaling="root"),
                preprocessing.scale_to_unit_norm,
            ]

        # Create the file if it doesn't exist.
        if not self.path.exists() or self.overwrite:
            with h5py.File(self.path, "w") as index:
                index.attrs["ms_level"] = self.ms_level
                index.attrs["n_spectra"] = 0
                index.attrs["n_peaks"] = 0
                index.attrs["annotated"] = self.annotated
                index.attrs["preprocessing_fn"] = _hash_obj(
                    tuple(self.preprocessing_fn)
                )
        else:
            self._validate_index()

        # Now parse spectra.
        if ms_data_files is not None:
            ms_data_files = utils.listify(ms_data_files)
            for ms_file in ms_data_files:
                self.add_file(ms_file)

    def _reindex(self) -> None:
        """Update the file mappings and offsets."""
        offsets = [0]
        for idx in range(len(self._handle)):
            grp = self._handle[str(idx)]
            offsets.append(grp.attrs["n_spectra"])
            self._file_map[grp.attrs["path"]] = idx

        self._file_offsets = np.cumsum([0] + offsets)

        # Build a map of 1D indices to 2D locations:
        grp_idx = 0
        for lin_idx in range(offsets[-1]):
            grp_idx += lin_idx >= offsets[grp_idx + 1]
            row_idx = lin_idx - offsets[grp_idx]
            self._locs[lin_idx] = (grp_idx, row_idx)

        self._offsets = None
        _ = self.offsets  # Reinitialize the offsets.

    def _validate_index(self) -> None:
        """Validate that the index is appropriate for this dataset."""
        preproc_hash = _hash_obj(tuple(self.preprocessing_fn))
        with self:
            try:
                assert self._handle.attrs["ms_level"] == self.ms_level
                assert self._handle.attrs["preprocessing_fn"] == preproc_hash
                if self._annotated:
                    assert self._handle.attrs["annotated"]
            except (KeyError, AssertionError):
                raise ValueError(
                    f"'{self.path}' already exists, but was created with "
                    "incompatible parameters. Use 'overwrite=True' to "
                    "overwrite it."
                )

            self._reindex()

    def _get_parser(
        self,
        ms_data_file: PathLike,
    ) -> MzmlParser | MzxmlParser | MgfParser:
        """Get the parser for the MS data file.

        Parameters
        ----------
        ms_data_file : PathLike
            The mass spectrometry data file to be parsed.

        Returns
        -------
        MzmlParser, MzxmlParser, or MgfParser
            The appropriate parser for the file.
        """
        kw_args = {
            "ms_level": self.ms_level,
            "valid_charge": self.valid_charge,
            "preprocessing_fn": self.preprocessing_fn,
        }

        if ms_data_file.suffix.lower() == ".mzml":
            return MzmlParser(ms_data_file, **kw_args)

        if ms_data_file.suffix.lower() == ".mzxml":
            return MzxmlParser(ms_data_file, **kw_args)

        if ms_data_file.suffix.lower() == ".mgf":
            return MgfParser(ms_data_file, **kw_args)

        raise ValueError("Only mzML, mzXML, and MGF files are supported.")

    def _assemble_metadata(
        self,
        parser: MzmlParser | MzxmlParser | MgfParser,
    ) -> np.ndarray:
        """Assemble the metadata.

        Parameters
        ----------
        parser : MzmlParser, MzxmlParser, MgfParser
            The parser to use.

        Returns
        -------
        numpy.ndarray of shape (n_spectra,)
            The file metadata.
        """
        meta_types = [
            ("precursor_mz", np.float32),
            ("precursor_charge", np.uint8),
            ("offset", np.uint64),
            ("scan_id", np.uint32),
            ("global_rt", np.float32),
        ]

        metadata = np.empty(parser.n_spectra, dtype=meta_types)
        metadata["precursor_mz"] = parser.precursor_mz
        metadata["precursor_charge"] = parser.precursor_charge
        metadata["offset"] = parser.offset
        metadata["scan_id"] = parser.scan_id
        metadata["global_rt"] = parser.global_rts
        return metadata

    def add_file(self, ms_data_file: PathLike) -> None:
        """Add a mass spectrometry data file to the index.

        Parameters
        ----------
        ms_data_file : PathLike
            The mass spectrometry data file to add. It must be in an mzML or
            MGF file format and use an ``.mzML``, ``.mzXML``, or ``.mgf`` file
            extension.
        """
        ms_data_file = Path(ms_data_file)
        if str(ms_data_file) in self._file_map:
            return

        # Invalidate current offsets:
        self._offsets = None

        # Read the file:
        parser = self._get_parser(ms_data_file)
        parser.read()

        # Create the tables:
        metadata = self._assemble_metadata(parser)

        spectrum_types = [
            ("mz_array", np.float64),
            ("intensity_array", np.float32),
            ("rt_array", np.float32),
            ("level_array", np.uint8),
            ("fragment_array", np.uint8),
        ]

        spectra = np.zeros(parser.n_peaks, dtype=spectrum_types)
        spectra["mz_array"] = parser.mz_arrays
        spectra["intensity_array"] = parser.intensity_arrays
        spectra["rt_array"] = parser.rt_arrays
        spectra["level_array"] = parser.level_arrays
        spectra["fragment_array"] = parser.fragment_arrays

        # Write the tables:
        with h5py.File(self.path, "a") as index:
            group_index = len(index)
            group = index.create_group(str(group_index))
            group.attrs["path"] = str(ms_data_file)
            group.attrs["n_spectra"] = parser.n_spectra
            group.attrs["n_peaks"] = parser.n_peaks
            group.attrs["id_type"] = parser.id_type

            # Update overall stats:
            index.attrs["n_spectra"] += parser.n_spectra
            index.attrs["n_peaks"] += parser.n_peaks

            # Add the datasets:
            group.create_dataset(
                "metadata",
                data=metadata,
            )

            group.create_dataset(
                "spectra",
                data=spectra,
            )

            try:
                group.create_dataset(
                    "annotations",
                    data=parser.annotations,
                    dtype=h5py.string_dtype(),
                )
            except (KeyError, AttributeError, TypeError):
                pass

            self._file_map[str(ms_data_file)] = group_index
            end_offset = self._file_offsets[-1] + parser.n_spectra
            self._file_offsets = np.append(self._file_offsets, [end_offset])

            # Update the locations:
            grp_idx = len(self._file_offsets) - 2
            for row_idx in range(parser.n_spectra):
                lin_idx = row_idx + self._file_offsets[-2]
                self._locs[lin_idx] = (grp_idx, row_idx)

    def get_spectrum(self, idx: int) -> MassSpectrum:
        """Access a mass spectrum.

        Parameters
        ----------
        idx : int
            The index of the index of the mass spectrum to look-up.

        Returns
        -------
        tuple of numpy.ndarray
            The m/z values, intensity values, precurosr m/z, precurosr charge,
            and the spectrum annotation.
        """
        group_index, row_index = self._locs[idx]
        if self._handle is None:
            raise RuntimeError("Use the context manager for access.")

        grp = self._handle[str(group_index)]
        metadata = grp["metadata"]
        spectra = grp["spectra"]
        offsets = self.offsets[str(group_index)][row_index : row_index + 2]

        start_offset = offsets[0]
        if offsets.shape[0] == 2:
            stop_offset = offsets[1]
        else:
            stop_offset = spectra.shape[0]

        spectrum = spectra[start_offset:stop_offset]
        precursor = metadata[row_index]
        return MassSpectrum(
            filename=grp.attrs["path"],
            scan_id=f"{grp.attrs['id_type']}={metadata[row_index]['scan_id']}",
            mz=np.array(spectrum["mz_array"]),
            intensity=np.array(spectrum["intensity_array"]),
            rt=np.array(spectrum["rt_array"]),
            level=np.array(spectrum["level_array"]),
            fragment_label=np.array(spectrum["fragment_array"]),
            precursor_mz=precursor["precursor_mz"],
            precursor_charge=precursor["precursor_charge"],
            global_rt = np.array(precursor["global_rt"])
        )

    def get_spectrum_id(self, idx: int) -> tuple[str, str]:
        """Get the identifier for a mass spectrum.

        Parameters
        ----------
        idx : int
            The index of the mass spectrum in the SpectrumIndex.

        Returns
        -------
        ms_data_file : str
            The mass spectrometry data file from which the mass spectrum was
            originally parsed.
        identifier : str
            The mass spectrum identifier, per PSI recommendations.
        """
        group_index, row_index = self._locs[idx]
        if self._handle is None:
            raise RuntimeError("Use the context manager for access.")

        grp = self._handle[str(group_index)]
        ms_data_file = grp.attrs["path"]
        identifier = grp["metadata"][row_index]["scan_id"]
        prefix = grp.attrs["id_type"]
        return ms_data_file, f"{prefix}={identifier}"

    def loader(self, *args: tuple, **kwargs: dict) -> DataLoader:
        """A PyTorch DataLoader for the mass spectra.

        Parameters
        ----------
        *args : tuple
            Arguments passed initialize a torch.utils.data.DataLoader,
            excluding ``dataset`` and ``collate_fn``.
        **kwargs : dict
            Keyword arguments passed initialize a torch.utils.data.DataLoader,
            excluding ``dataset`` and ``collate_fn``.

        Returns
        -------
        torch.utils.data.DataLoader
            A DataLoader for the mass spectra.
        """
        return DataLoader(self, *args, collate_fn=self.collate_fn, **kwargs)

    def __len__(self) -> int:
        """The number of spectra in the index."""
        return self.n_spectra

    def __del__(self) -> None:
        """Cleanup the temporary directory."""
        if self._tmpdir is not None:
            self._tmpdir.cleanup()

    def __getitem__(self, idx: int) -> MassSpectrum:
        """Access a mass spectrum.

        Parameters
        ----------
        idx : int
            The overall index of the mass spectrum to retrieve.

        Returns
        -------
        tuple of numpy.ndarray
            The m/z values, intensity values, precursor m/z, precurosr charge,
            and the annotation (if available).
        """
        if self._handle is None:
            with self:
                return self.get_spectrum(idx)

        return self.get_spectrum(idx)

    def __enter__(self) -> SpectrumDataset:
        """Open the index file for reading."""
        if self._handle is None:
            self._handle = h5py.File(
                self.path,
                "r",
                rdcc_nbytes=int(3e8),
                rdcc_nslots=1024000,
            )
        return self

    def __exit__(self, *args: str) -> None:
        """Close the HDF5 file."""
        self._handle.close()
        self._handle = None

    @property
    def ms_files(self) -> list[str]:
        """The files currently in the index."""
        return list(self._file_map.keys())

    @property
    def path(self) -> Path:
        """The path to the underyling HDF5 index file."""
        return self._path

    @property
    def ms_level(self) -> int:
        """The MS level of tandem mass spectra in the collection."""
        return self._ms_level

    @property
    def preprocessing_fn(self) -> list[Callable]:
        """The functions for preprocessing MS data."""
        return self._preprocessing_fn

    @property
    def valid_charge(self) -> list[int]:
        """Valid precursor charges for spectra to be included."""
        return self._valid_charge

    @property
    def annotated(self) -> bool:
        """Whether or not the index contains spectrum annotations."""
        return self._annotated

    @property
    def overwrite(self) -> bool:
        """Overwrite a previous index?."""
        return self._overwrite

    @property
    def n_spectra(self) -> int:
        """The total number of mass spectra in the index."""
        if self._handle is None:
            with self:
                return self._handle.attrs["n_spectra"]

        return self._handle.attrs["n_spectra"]

    @property
    def n_peaks(self) -> int:
        """The total number of mass peaks in the index."""
        if self._handle is None:
            with self:
                return self._handle.attrs["n_peaks"]

        return self._handle.attrs["n_peaks"]

    @property
    def offsets(self) -> dict[str, np.array]:
        """The offsets denoting where each spectrum starts."""
        if self._offsets is not None:
            return self._offsets

        self._offsets = {
            k: v["metadata"]["offset"] for k, v in self._handle.items()
        }

        return self._offsets

    @staticmethod
    def collate_fn(
        batch: Iterable[MassSpectrum],
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """The collate function for a SpectrumDataset.

        The mass spectra must be padded so that they fit nicely as a tensor.
        However, the padded elements are ignored during the subsequent steps.

        Parameters
        ----------
        batch : tuple of tuple of torch.Tensor
            A batch of data from an SpectrumDataset.

        Returns
        -------
        spectra : torch.Tensor of shape (batch_size, n_peaks, 2)
            The mass spectra to sequence, where ``X[:, :, 0]`` are the m/z
            values and ``X[:, :, 1]`` are their associated intensities.
        precursors : torch.Tensor of shape (batch_size, 2)
            The precursor mass and charge state.
        """
        return _collate_fn(batch)[0:2]


class AnnotatedSpectrumDataset(SpectrumDataset):
    """Store and access a collection of annotated mass spectra.

    This class parses one or more MGF file and converts it to our HDF5 index
    format. This allows us to access spectra from many different files quickly
    and without loading them all into memory.

    Parameters
    ----------
    index_path : str
        The name and path of the HDF5 file index. If the path does
        not contain the `.h5` or `.hdf5` extension, `.hdf5` will be added.
    ms_data_files : str or list of str, optional
        The MGF to include in this collection.
    ms_level : int, optional
        The level of tandem mass spectra to use.
    preprocessing_fn : Callable or Iterable[Callable], optional
        The function(s) used to preprocess the mass spectra. ``None``,
        the default, filters for the top 200 peaks above m/z 140,
        square root transforms the intensities and scales them to unit norm.
        See the preprocessing module for details and additional options.
    valid_charge : Iterable[int], optional
        Only consider spectra with the specified precursor charges. If `None`,
        any precursor charge is accepted.
    overwrite : bool
        Overwrite previously indexed files? If ``False`` and new files are
        provided, they will be appended to the collection.

    Attributes
    ----------
    ms_files : list of str
    path : Path
    ms_level : int
    valid_charge : Optional[Iterable[int]]
    overwrite : bool
    n_spectra : int
    n_peaks : int
    """

    _annotated = True

    def __init__(
        self,
        tokenizer: PeptideTokenizer,
        ms_data_files: PathLike | Iterable[PathLike] = None,
        ms_level: int = 2,
        preprocessing_fn: Callable | Iterable[Callable] | None = None,
        valid_charge: Iterable[int] | None = None,
        index_path: PathLike | None = None,
        overwrite: bool = False,
    ) -> None:
        """Initialize an AnnotatedSpectrumIndex."""
        self.tokenizer = tokenizer
        super().__init__(
            ms_data_files=ms_data_files,
            ms_level=ms_level,
            preprocessing_fn=preprocessing_fn,
            valid_charge=valid_charge,
            index_path=index_path,
            overwrite=overwrite,
        )

    def _get_parser(self, ms_data_file: str) -> MgfParser:
        """Get the parser for the MS data file."""
        if ms_data_file.suffix.lower() == ".mgf":
            return MgfParser(
                ms_data_file,
                ms_level=self.ms_level,
                annotations=True,
                preprocessing_fn=self.preprocessing_fn,
            )
        elif ms_data_file.suffix.lower() == ".asf":
            return AsfParser(
                ms_data_file,
                ms_level=self.ms_level,
                annotations=True,
                preprocessing_fn=self.preprocessing_fn,
            )

        raise ValueError("Only MGF files are currently supported.")

    def get_spectrum(self, idx: int) -> MassSpectrum:
        """Access a mass spectrum.

        Parameters
        ----------
        idx : int
            The index of the mass spectrum in the AnnotatedSpectrumIndex.

        Returns
        -------
        MassSpectrum
            The mass spectrum, labeled with its annotation.
        """
        spectrum = super().get_spectrum(idx)
        group_index, row_index = self._locs[idx]
        grp = self._handle[str(group_index)]
        annotations = grp["annotations"]
        spectrum.label = annotations[row_index].decode()
        return spectrum

    @property
    def annotations(self) -> np.ndarray[str]:
        """Retrieve all of the annotations in the index."""
        annotations = []
        for grp in self._handle.values():
            try:
                annotations.append(grp["annotations"])
            except KeyError:
                pass

        return np.concatenate(annotations)

    def collate_fn(
        self,
        batch: Iterable[MassSpectrum],
    ) -> tuple[torch.Tensor, torch.Tensor, np.ndarray[str]]:
        """The collate function for a AnnotatedSpectrumDataset.

        The mass spectra must be padded so that they fit nicely as a tensor.
        However, the padded elements are ignored during the subsequent steps.

        Parameters
        ----------
        batch : tuple of tuple of torch.Tensor
            A batch of data from an AnnotatedSpectrumDataset.

        Returns
        -------
        spectra : torch.Tensor of shape (batch_size, n_peaks, 2)
            The mass spectra to sequence, where ``X[:, :, 0]`` are the m/z
            values and ``X[:, :, 1]`` are their associated intensities.
        precursors : torch.Tensor of shape (batch_size, 2)
            The precursor mass and charge state.
        annotations : np.ndarray[str]
            The spectrum annotations.
        """
        spectra, precursors, annotations, fragment_labels, global_rt = _collate_fn(batch)
        tokens = self.tokenizer.tokenize(annotations, add_stop=True)
        return spectra, precursors, tokens, fragment_labels, global_rt


def _collate_fn(
    batch: Iterable[MassSpectrum],
) -> tuple[torch.Tensor, torch.Tensor, list[str | None]]:
    """The collate function for a SpectrumDataset.

    The mass spectra must be padded so that they fit nicely as a tensor.
    However, the padded elements are ignored during the subsequent steps.

    Parameters
    ----------
    batch : tuple of tuple of torch.Tensor
        A batch of data from an SpectrumDataset.

    Returns
    -------
    spectra : torch.Tensor of shape (batch_size, n_peaks, 2)
        The mass spectra to sequence, where ``X[:, :, 0]`` are the m/z
        values and ``X[:, :, 1]`` are their associated intensities.
    precursors : torch.Tensor of shape (batch_size, 2)
        The precursor mass and charge state.
    annotations : np.ndarray of shape (batch_size,)
        The annotations, if any exist.
    """
    spectra = []
    masses = []
    charges = []
    annotations = []
    fragment_label = []
    global_rt = []
    for spec in batch:
        spectra.append(spec.to_tensor())
        masses.append(spec.precursor_mass) #spec.precursor_mass + np.random.rand() * spec.precursor_mass * 50 / 1000000)
        charges.append(spec.precursor_charge)
        annotations.append(spec.label)
        fragment_label.append(torch.tensor(spec.fragment_label))
        global_rt.append(spec.global_rt)
    precursor_masses = (np.array(masses) - 1.007276) * np.array(charges)
    precursors = torch.vstack([torch.tensor(precursor_masses), torch.tensor(charges)])
    spectra = torch.nn.utils.rnn.pad_sequence(
        spectra,
        batch_first=True,
    )
    fragment_label = torch.nn.utils.rnn.pad_sequence(
        fragment_label,
        batch_first=True,
    )

    model_type = torch.float32
    return spectra.type(model_type), precursors.T.type(model_type), annotations, fragment_label.flatten(), global_rt


def _hash_obj(obj: Any) -> str:  # noqa: ANN401
    """SHA1 hash for a picklable object."""
    out = hashlib.sha1()
    out.update(dill.dumps(obj))
    return out.hexdigest()



"""Simple encoders for input into Transformers and the like."""
import math

import einops
import numpy as np
import torch


class FloatEncoder(torch.nn.Module):
    """Encode floating point values using sine and cosine waves.

    Parameters
    ----------
    d_model : int
        The number of features to output.
    min_wavelength : float, optional
        The minimum wavelength to use.
    max_wavelength : float, optional
        The maximum wavelength to use.
    learnable_wavelengths : bool, optional
        Allow the selected wavelengths to be fine-tuned
        by the model.
    """

    def __init__(
        self,
        d_model: int,
        min_wavelength: float = 0.001,
        max_wavelength: float = 10000,
        learnable_wavelengths: bool = False,
    ) -> None:
        """Initialize the MassEncoder."""
        super().__init__()

        # Error checking:
        if min_wavelength <= 0:
            raise ValueError("'min_wavelength' must be greater than 0.")

        if max_wavelength <= 0:
            raise ValueError("'max_wavelength' must be greater than 0.")

        self.learnable_wavelengths = learnable_wavelengths

        # Get dimensions for equations:
        d_sin = math.ceil(d_model / 2)
        d_cos = d_model - d_sin

        base = min_wavelength / (2 * np.pi)
        scale = max_wavelength / min_wavelength
        sin_exp = torch.arange(0, d_sin).float() / (d_sin - 1)
        cos_exp = (torch.arange(d_sin, d_model).float() - d_sin) / (d_cos - 1)
        sin_term = base * (scale**sin_exp)
        cos_term = base * (scale**cos_exp)

        if not self.learnable_wavelengths:
            self.register_buffer("sin_term", sin_term)
            self.register_buffer("cos_term", cos_term)
        else:
            self.sin_term = torch.nn.Parameter(sin_term)
            self.cos_term = torch.nn.Parameter(cos_term)

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """Encode m/z values.

        Parameters
        ----------
        X : torch.Tensor of shape (batch_size, n_float)
            The masses to embed.

        Returns
        -------
        torch.Tensor of shape (batch_size, n_float, d_model)
            The encoded features for the floating point numbers.
        """
        sin_mz = torch.sin(X[:, :, None] / self.sin_term)
        cos_mz = torch.cos(X[:, :, None] / self.cos_term)
        return torch.cat([sin_mz, cos_mz], axis=-1)


class PeakEncoder(torch.nn.Module):
    """Encode mass spectrum.

    Parameters
    ----------
    d_model : int
        The number of features to output.
    min_mz_wavelength : float, optional
        The minimum wavelength to use for m/z.
    max_mz_wavelength : float, optional
        The maximum wavelength to use for m/z.
    min_intensity_wavelength : float, optional
        The minimum wavelength to use for intensity. The default assumes
        intensities between [0, 1].
    max_intensity_wavelength : float, optional
        The maximum wavelength to use for intensity. The default assumes
        intensities between [0, 1].
    learnable_wavelengths : bool, optional
        Allow the selected wavelengths to be fine-tuned
        by the model.
    """

    def __init__(
        self,
        d_model: int,
        min_mz_wavelength: float = 0.001,
        max_mz_wavelength: float = 10000,
        min_intensity_wavelength: float = 1e-6,
        max_intensity_wavelength: float = 1,
        learnable_wavelengths: bool = False,
    ) -> None:
        """Initialize the MzEncoder."""
        super().__init__()
        self.d_model = d_model
        self.learnable_wavelengths = learnable_wavelengths

        self.mz_encoder = FloatEncoder(
            d_model=self.d_model,
            min_wavelength=min_mz_wavelength,
            max_wavelength=max_mz_wavelength,
            learnable_wavelengths=learnable_wavelengths,
        )

        self.int_encoder = FloatEncoder(
            d_model=self.d_model,
            min_wavelength=min_intensity_wavelength,
            max_wavelength=max_intensity_wavelength,
            learnable_wavelengths=learnable_wavelengths,
        )
        self.combiner = torch.nn.Linear(2 * d_model, d_model, bias=False)

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """Encode m/z values and intensities.

        Note that we expect intensities to fall within the interval [0, 1].

        Parameters
        ----------
        X : torch.Tensor of shape (n_spectra, n_peaks, 2)
            The spectra to embed. Axis 0 represents a mass spectrum, axis 1
            contains the peaks in the mass spectrum, and axis 2 is essentially
            a 2-tuple specifying the m/z-intensity pair for each peak. These
            should be zero-padded, such that all of the spectra in the batch
            are the same length.

        Returns
        -------
        torch.Tensor of shape (n_spectra, n_peaks, d_model)
            The encoded features for the mass spectra.
        """
        encoded = torch.cat(
            [
                self.mz_encoder(X[:, :, 0]),
                self.int_encoder(X[:, :, 1]),
            ],
            dim=2,
        )

        return self.combiner(encoded)


class PositionalEncoder(FloatEncoder):
    """The positional encoder for sequences.

    Parameters
    ----------
    d_model : int
        The number of features to output.
    min_wavelength : float, optional
        The shortest wavelength in the geometric progression.
    max_wavelength : float, optional
        The longest wavelength in the geometric progression.
    """

    def __init__(
        self,
        d_model: int,
        min_wavelength: float = 1.0,
        max_wavelength: float = 1e5,
    ) -> None:
        """Initialize the MzEncoder."""
        super().__init__(
            d_model=d_model,
            min_wavelength=min_wavelength,
            max_wavelength=max_wavelength,
        )

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """Encode positions in a sequence.

        Parameters
        ----------
        X : torch.Tensor of shape (batch_size, n_sequence, n_features)
            The first dimension should be the batch size (i.e. each is one
            peptide) and the second dimension should be the sequence (i.e.
            each should be an amino acid representation).

        Returns
        -------
        torch.Tensor of shape (batch_size, n_sequence, n_features)
            The encoded features for the mass spectra.
        """
        pos = torch.arange(X.shape[1]).type_as(self.sin_term)
        pos = einops.repeat(pos, "n -> b n", b=X.shape[0])
        sin_in = einops.repeat(pos, "b n -> b n f", f=len(self.sin_term))
        cos_in = einops.repeat(pos, "b n -> b n f", f=len(self.cos_term))

        sin_pos = torch.sin(sin_in / self.sin_term)
        cos_pos = torch.cos(cos_in / self.cos_term)
        encoded = torch.cat([sin_pos, cos_pos], axis=2)
        return encoded + X



"""Transformer models for peptides."""
import torch

from ..encoders import FloatEncoder, PositionalEncoder
from ..tokenizers import PeptideTokenizer


class _PeptideTransformer(torch.nn.Module):
    """A transformer base class for peptide sequences.

    Parameters
    ----------
    n_tokens : int or PeptideTokenizer
        The number of tokens used to tokenize peptide sequences.
    d_model : int
        The latent dimensionality to represent the amino acids in a peptide
        sequence.
    positional_encoder : PositionalEncoder or bool
        The positional encodings to use for the amino acid sequence. If
        ``True``, the default positional encoder is used. ``False`` disables
        positional encodings, typically only for ablation tests.
    max_charge : int
        The maximum precursor charge to embed.
    """

    def __init__(
        self,
        n_tokens: int | PeptideTokenizer,
        d_model: int,
        positional_encoder: PositionalEncoder | bool,
        max_charge: int,
    ) -> None:
        super().__init__()
        try:
            n_tokens = len(n_tokens)
        except TypeError:
            pass

        if callable(positional_encoder):
            self.positional_encoder = positional_encoder
        elif positional_encoder:
            self.positional_encoder = PositionalEncoder(d_model)
        else:
            self.positional_encoder = torch.nn.Identity()

        self.charge_encoder = torch.nn.Embedding(max_charge + 1, d_model)
        self.aa_encoder = torch.nn.Embedding(
            n_tokens + 1,
            d_model,
            padding_idx=0,
        )

    @property
    def device(self) -> torch.device:
        """The current device for the model."""
        return next(self.parameters()).device


class PeptideTransformerEncoder(_PeptideTransformer):
    """A transformer encoder for peptide sequences.

    Parameters
    ----------
    n_tokens : int or PeptideTokenizer
        The number of tokens used to tokenize peptide sequences.
    d_model : int
        The latent dimensionality to represent the amino acids in a peptide
        sequence.
    nhead : int, optional
        The number of attention heads in each layer. ``d_model`` must be
        divisible by ``nhead``.
    dim_feedforward : int, optional
        The dimensionality of the fully connected layers in the Transformer
        layers of the model.
    n_layers : int, optional
        The number of Transformer layers.
    dropout : float, optional
        The dropout probability for all layers.
    positional_encoder : PositionalEncoder or bool, optional
        The positional encodings to use for the amino acid sequence. If
        ``True``, the default positional encoder is used. ``False`` disables
        positional encodings, typically only for ablation tests.
    max_charge : int, optional
        The maximum charge state for peptide sequences.
    """

    def __init__(
        self,
        n_tokens: int | PeptideTokenizer,
        d_model: int = 128,
        nhead: int = 8,
        dim_feedforward: int = 1024,
        n_layers: int = 1,
        dropout: float = 0,
        positional_encoder: PositionalEncoder | bool = True,
        max_charge: int = 5,
    ) -> None:
        """Initialize a PeptideEncoder."""
        super().__init__(
            n_tokens=n_tokens,
            d_model=d_model,
            positional_encoder=positional_encoder,
            max_charge=max_charge,
        )

        # The Transformer layers:
        layer = torch.nn.TransformerEncoderLayer(
            d_model=d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            batch_first=True,
            dropout=dropout,
        )

        self.transformer_encoder = torch.nn.TransformerEncoder(
            layer,
            num_layers=n_layers,
        )

    def forward(
        self,
        tokens: torch.Tensor,
        charges: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Predict the next amino acid for a collection of sequences.

        Parameters
        ----------
        tokens : torch.Tensor of size (batch_size, peptide_length)
            The integer tokens describing each peptide sequence, padded
            to the maximum peptide length in the batch with 0s.
        charges : torch.Tensor of size (batch_size,)
            The charge state of each peptide.

        Returns
        -------
        latent : torch.Tensor of shape (n_sequences, len_sequence, d_model)
            The latent representations for the spectrum and each of its
            peaks.
        mem_mask : torch.Tensor
            The memory mask specifying which elements were padding in X.
        """
        # Encode everything:
        encoded = self.aa_encoder(tokens)
        charges = self.charge_encoder(charges)[:, None]
        encoded = torch.cat([charges, encoded], dim=1)

        # Create mask
        mask = ~encoded.sum(dim=2).bool()

        # Add positional encodings
        encoded = self.positional_encoder(encoded)

        # Run through the model:
        latent = self.transformer_encoder(encoded, src_key_padding_mask=mask)
        return latent, mask


class PeptideTransformerDecoder(_PeptideTransformer):
    """A transformer decoder for peptide sequences.

    Parameters
    ----------
    n_tokens : int or PeptideTokenizer
        The number of tokens used to tokenize peptide sequences.
    d_model : int, optional
        The latent dimensionality to represent peaks in the mass spectrum.
    nhead : int, optional
        The number of attention heads in each layer. ``d_model`` must be
        divisible by ``nhead``.
    dim_feedforward : int, optional
        The dimensionality of the fully connected layers in the Transformer
        layers of the model.
    n_layers : int, optional
        The number of Transformer layers.
    dropout : float, optional
        The dropout probability for all layers.
    positional_encoder : PositionalEncoder or bool, optional
        The positional encodings to use for the amino acid sequence. If
        ``True``, the default positional encoder is used. ``False`` disables
        positional encodings, typically only for ablation tests.
    max_charge : int, optional
        The maximum charge state for peptide sequences.
    """

    def __init__(
        self,
        n_tokens: int | PeptideTokenizer,
        d_model: int = 128,
        nhead: int = 8,
        dim_feedforward: int = 1024,
        n_layers: int = 1,
        dropout: float = 0,
        positional_encoder: PositionalEncoder | bool = True,
        max_charge: int = 5,
    ) -> None:
        """Initialize a PeptideDecoder."""
        super().__init__(
            n_tokens=n_tokens,
            d_model=d_model,
            positional_encoder=positional_encoder,
            max_charge=max_charge,
        )

        # Additional model components
        self.mass_encoder = FloatEncoder(d_model)
        layer = torch.nn.TransformerDecoderLayer(
            d_model=d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            batch_first=True,
            dropout=dropout,
        )

        self.transformer_decoder = torch.nn.TransformerDecoder(
            layer,
            num_layers=n_layers,
        )

        self.final = torch.nn.Linear(
            d_model,
            self.aa_encoder.num_embeddings,
        )

    def forward(
        self,
        tokens: torch.Tensor | None,
        precursors: torch.Tensor,
        memory: torch.Tensor,
        memory_key_padding_mask: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Predict the next amino acid for a collection of sequences.

        Parameters
        ----------
        tokens : list of str, torch.Tensor, or None
            The partial peptide sequences for which to predict the next
            amino acid. Optionally, these may be the token indices instead
            of a string.
        precursors : torch.Tensor of size (batch_size, 2)
            The precursor mass (axis 0) and charge (axis 1).
        memory : torch.Tensor of shape (batch_size, n_peaks, d_model)
            The representations from a ``TransformerEncoder``, such as a
            ``SpectrumEncoder``.
        memory_key_padding_mask : torch.Tensor of shape (batch_size, n_peaks)
            The mask that indicates which elements of ``memory`` are padding.

        Returns
        -------
        scores : torch.Tensor of size (batch_size, len_sequence, n_amino_acids)
            The raw output for the final linear layer. These can be Softmax
            transformed to yield the probability of each amino acid for the
            prediction.

        """
        # Prepare sequences
        if tokens is None:
            tokens = torch.tensor([[]]).to(self.device)

        # Encode everything:
        tokens = self.aa_encoder(tokens)
        masses = self.mass_encoder(precursors[:, None, 0])
        charges = self.charge_encoder(precursors[:, 1].int() - 1)
        precursors = masses + charges[:, None, :]

        # Feed through model:
        tgt = torch.cat([precursors, tokens], dim=1)
        tgt_key_padding_mask = tgt.sum(axis=2) == 0
        tgt = self.positional_encoder(tgt)
        tgt_mask = generate_tgt_mask(tgt.shape[1]).to(self.device)
        preds = self.transformer_decoder(
            tgt=tgt,
            memory=memory,
            tgt_mask=tgt_mask,
            tgt_key_padding_mask=tgt_key_padding_mask,
            memory_key_padding_mask=memory_key_padding_mask.to(self.device),
        )
        return self.final(preds)


def generate_tgt_mask(sz: int) -> torch.Tensor:
    """Generate a square mask for the sequence.

    Parameters
    ----------
    sz : int
        The length of the target sequence.
    """
    return ~torch.triu(torch.ones(sz, sz, dtype=torch.bool)).transpose(0, 1)



from .depthcharge.tokenizers.peptides import PeptideTokenizer, Peptide,  _calc_precursor_mass, nb
import numpy as np
from pyteomics import mass
import re

def convert_mods_to_shifts(db, pep):
    toks = re.split(r'(\[|\])', pep)
    converted_toks = []
    ii = 0
    while ii < len(toks):
        converted_toks.append(toks[ii])
        if toks[ii] == '[':
            try:
                mod_mass = db.by_name(toks[ii+1]).monoisotopic_mass
                if mod_mass < 0:
                    mod_mass = str(round(mod_mass,2))
                else:
                     mod_mass = '+' + str(round(mod_mass,2))
                converted_toks.append(mod_mass)
                ii += 1
            except(KeyError):
                pass
        ii += 1
    print(converted_toks)
    return(''.join(converted_toks))

def write_results(preds, results_file, raw_file, isolation_window_size, score_threshold, time_width):
    pred_seqs, aa_conf, pep_conf, rts, precursors = [], [], [], [], []

    unimod_db = mass.unimod.Unimod()

    for pred_seqs_c, _, pep_conf_c, aa_conf_c, rt_c, precursors_c in preds:
        pred_seqs.append(pred_seqs_c)
        aa_conf.append(aa_conf_c)
        pep_conf.append(pep_conf_c)
        rts.append(rt_c)
        precursors.append(precursors_c)
        
    pred_seqs = np.concatenate(pred_seqs)
    pep_conf = np.concatenate(pep_conf)
    rts = np.concatenate(rts)
    precursors = np.concatenate(precursors)

    if max(rts) > 500:
      rts = rts/60
      time_width = time_width/60

    tokenizer = PeptideTokenizer.from_massivekb(reverse=False, replace_isoleucine_with_leucine=True)
    with open(results_file + '.ssl', 'w') as out:
        out.write("\t".join(['file',	'scan',	'charge',	'sequence',	'score-type',	'score',	'retention-time',	'start-time',	'end-time']) + '\n')
        for idx, (pred, conf, rt, prec) in enumerate(zip(pred_seqs, pep_conf, rts, precursors)):
            conf = np.exp(conf)
            unmodified = re.sub("\[.*?\]","", pred)
            if '-' in pred:
                if unmodified.rfind('-') != 0:
                    continue
            if len(unmodified) > 0 and conf > score_threshold:
              mz = prec[0]
              charge = prec[1]
              mz = mz / charge + 1.007276
              mz = mz / charge + 1.007276
              pred_mz = _calc_precursor_mass(nb.typed.List(Peptide.from_massivekb(pred).split()), charge, tokenizer.residues)
              pred = convert_mods_to_shifts(unimod_db, pred)
              if np.abs(pred_mz - mz) < 2*isolation_window_size:
                out.write("\t".join([raw_file, str(idx), str(int(charge)), pred, 'UNKNOWN', str(conf), str(rt), str(rt-time_width), str(rt+time_width)]) + '\n')